# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the dialogue/base.py module."""

import re
from typing import FrozenSet, Tuple, Type, cast
from unittest.mock import Mock, patch

import pytest

from aea.common import Address
from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError
from aea.helpers.storage.generic_storage import Storage
from aea.protocols.base import Message
from aea.protocols.dialogue.base import (
    BasicDialoguesStorage,
)
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import (
    DialogueLabel,
    DialogueMessage,
    DialogueStats,
)
from aea.protocols.dialogue.base import Dialogues as BaseDialogues
from aea.protocols.dialogue.base import (
    InvalidDialogueMessage,
    PersistDialoguesStorage,
    PersistDialoguesStorageWithOffloading,
    find_caller_object,
)
from aea.skills.base import SkillComponent

from packages.fetchai.protocols.default.custom_types import ErrorCode
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.state_update.message import StateUpdateMessage

from tests.common.utils import wait_for_condition


class Dialogue(BaseDialogue):
    """This concrete class defines a dialogue."""

    INITIAL_PERFORMATIVES = frozenset({DefaultMessage.Performative.BYTES})
    TERMINAL_PERFORMATIVES = frozenset({DefaultMessage.Performative.ERROR})
    VALID_REPLIES = {
        DefaultMessage.Performative.BYTES: frozenset(
            {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
        ),
        DefaultMessage.Performative.ERROR: frozenset(),
    }

    class Role(BaseDialogue.Role):
        """This class defines the agent's role in this dialogue."""

        ROLE1 = "role1"
        ROLE2 = "role2"

    class EndState(BaseDialogue.EndState):
        """This class defines the end states of this dialogue."""

        SUCCESSFUL = 0
        FAILED = 1

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        message_class: Type[Message] = DefaultMessage,
        self_address: Address = "agent 1",
        role: BaseDialogue.Role = Role.ROLE1,
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :return: None
        """
        BaseDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class Dialogues(BaseDialogues):
    """This class gives a concrete definition of dialogues."""

    END_STATES = frozenset(
        {Dialogue.EndState.SUCCESSFUL, Dialogue.EndState.FAILED}
    )  # type: FrozenSet[BaseDialogue.EndState]

    def __init__(
        self,
        self_address: Address,
        message_class=DefaultMessage,
        dialogue_class=Dialogue,
        keep_terminal_state_dialogues=None,
    ) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom dialogues are maintained
        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return Dialogue.Role.ROLE1

        BaseDialogues.__init__(
            self,
            self_address=self_address,
            end_states=cast(FrozenSet[BaseDialogue.EndState], self.END_STATES),
            message_class=message_class,
            dialogue_class=dialogue_class,
            role_from_first_message=role_from_first_message,
            keep_terminal_state_dialogues=keep_terminal_state_dialogues,
        )


def test_dialogue_message_python():
    """Test DiallogueMessage."""
    dialogue_message = DialogueMessage(DefaultMessage.Performative.BYTES)
    assert isinstance(dialogue_message.performative, Message.Performative)

    assert dialogue_message.performative == DefaultMessage.Performative.BYTES
    assert dialogue_message.contents == {}
    assert dialogue_message.is_incoming is None
    assert dialogue_message.target is None


class TestDialogueLabel:
    """Test for DialogueLabel."""

    @classmethod
    def setup_class(cls):
        """Initialise the environment to test DialogueLabel."""
        cls.agent_address = "agent 1"
        cls.opponent_address = "agent 2"
        cls.dialogue_starter_ref = str(1)
        cls.dialogue_label = DialogueLabel(
            dialogue_reference=(
                cls.dialogue_starter_ref,
                Dialogue.UNASSIGNED_DIALOGUE_REFERENCE,
            ),
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.agent_address,
        )

    def test_all_methods(self):
        """Test the DialogueLabel."""
        assert self.dialogue_label.dialogue_reference == (
            self.dialogue_starter_ref,
            Dialogue.UNASSIGNED_DIALOGUE_REFERENCE,
        )
        assert (
            self.dialogue_label.dialogue_starter_reference == self.dialogue_starter_ref
        )
        assert (
            self.dialogue_label.dialogue_responder_reference
            == Dialogue.UNASSIGNED_DIALOGUE_REFERENCE
        )
        assert self.dialogue_label.dialogue_opponent_addr == self.opponent_address
        assert self.dialogue_label.dialogue_starter_addr == self.agent_address
        assert str(self.dialogue_label) == "{}_{}_{}_{}".format(
            self.dialogue_label.dialogue_starter_reference,
            self.dialogue_label.dialogue_responder_reference,
            self.dialogue_label.dialogue_opponent_addr,
            self.dialogue_label.dialogue_starter_addr,
        )

        dialogue_label_eq = DialogueLabel(
            dialogue_reference=(
                self.dialogue_starter_ref,
                Dialogue.UNASSIGNED_DIALOGUE_REFERENCE,
            ),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )

        assert dialogue_label_eq == self.dialogue_label

        dialogue_label_not_eq = "This is a test"

        assert not dialogue_label_not_eq == self.dialogue_label

        assert hash(dialogue_label_eq) == hash(self.dialogue_label)

        assert self.dialogue_label.json == dict(
            dialogue_starter_reference=self.dialogue_starter_ref,
            dialogue_responder_reference=Dialogue.UNASSIGNED_DIALOGUE_REFERENCE,
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        assert DialogueLabel.from_json(self.dialogue_label.json) == self.dialogue_label
        assert DialogueLabel.from_str(str(self.dialogue_label)) == self.dialogue_label
        assert not self.dialogue_label.is_complete()

        (
            incomplete_dialogue_label,
            complete_dialogue_label,
        ) = self.dialogue_label.get_both_versions()
        assert incomplete_dialogue_label.dialogue_reference == (
            self.dialogue_label.dialogue_starter_reference,
            Dialogue.UNASSIGNED_DIALOGUE_REFERENCE,
        )
        assert complete_dialogue_label is None


class TestDialogueBase:
    """Test for Dialogue."""

    def setup_method(self):
        """Initialise the environment to test Dialogue."""
        self.incomplete_reference = (str(1), "")
        self.complete_reference = (str(1), str(1))
        self.opponent_address = "agent 2"
        self.agent_address = "agent 1"

        self.dialogue_label = DialogueLabel(
            dialogue_reference=self.incomplete_reference,
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.dialogue = Dialogue(dialogue_label=self.dialogue_label)

        self.dialogue_label_opponent_started = DialogueLabel(
            dialogue_reference=self.complete_reference,
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.opponent_address,
        )
        self.dialogue_opponent_started = Dialogue(
            dialogue_label=self.dialogue_label_opponent_started
        )

        # convenient messages to reuse across tests
        self.valid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        self.valid_message_1_by_self.sender = self.agent_address
        self.valid_message_1_by_self.to = self.opponent_address

        self.valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=-1,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        self.valid_message_2_by_other.sender = self.opponent_address
        self.valid_message_2_by_other.to = self.agent_address

        self.valid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=-1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        self.valid_message_3_by_self.sender = self.agent_address
        self.valid_message_3_by_self.to = self.opponent_address

    def test_inner_classes(self):
        """Test the inner classes: Role and EndStates."""
        assert str(Dialogue.Role.ROLE1) == "role1"
        assert str(Dialogue.Role.ROLE2) == "role2"
        assert str(Dialogue.EndState.SUCCESSFUL) == "0"
        assert str(Dialogue.EndState.FAILED) == "1"

    def test_dialogue_properties(self):
        """Test dialogue properties."""
        assert self.dialogue.dialogue_label == self.dialogue_label
        assert self.dialogue.incomplete_dialogue_label == self.dialogue_label
        assert self.dialogue.dialogue_labels == {self.dialogue_label}
        assert self.dialogue.self_address == self.agent_address

        assert self.dialogue.role == Dialogue.Role.ROLE1
        assert str(self.dialogue.role) == "role1"

        assert self.dialogue.rules.initial_performatives == frozenset(
            {DefaultMessage.Performative.BYTES}
        )
        assert self.dialogue.rules.terminal_performatives == frozenset(
            {DefaultMessage.Performative.ERROR}
        )
        assert self.dialogue.rules.valid_replies == {
            DefaultMessage.Performative.BYTES: frozenset(
                {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
            ),
            DefaultMessage.Performative.ERROR: frozenset(),
        }
        assert self.dialogue.rules.get_valid_replies(
            DefaultMessage.Performative.BYTES
        ) == frozenset(
            {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
        )
        assert self.dialogue.rules.get_valid_replies(
            DefaultMessage.Performative.ERROR
        ) == frozenset({})
        assert self.dialogue.message_class == DefaultMessage

        assert self.dialogue.is_self_initiated

        assert self.dialogue.last_incoming_message is None
        assert self.dialogue.last_outgoing_message is None
        assert self.dialogue.last_message is None

        assert self.dialogue.is_empty

    def test_counterparty_from_message(self):
        """Test the 'counterparty_from_message' method."""
        assert (
            self.dialogue._counterparty_from_message(self.valid_message_1_by_self)
            == self.opponent_address
        )
        assert (
            self.dialogue._counterparty_from_message(self.valid_message_2_by_other)
            == self.opponent_address
        )

    def test_is_message_by_self(self):
        """Test the 'is_message_by_self' method."""
        assert self.dialogue._is_message_by_self(self.valid_message_1_by_self)
        assert not self.dialogue._is_message_by_self(self.valid_message_2_by_other)

    def test_is_message_by_other(self):
        """Test the 'is_message_by_other' method."""
        assert not self.dialogue._is_message_by_other(self.valid_message_1_by_self)
        assert self.dialogue._is_message_by_other(self.valid_message_2_by_other)

    def test_try_get_message(self):
        """Test the 'try_get_message' method."""
        assert (
            self.dialogue.get_message_by_id(self.valid_message_1_by_self.message_id)
            is None
        )
        self.dialogue._update(self.valid_message_1_by_self)
        assert (
            self.dialogue.get_message_by_id(self.valid_message_1_by_self.message_id)
            == self.valid_message_1_by_self
        )

        assert (
            self.dialogue.get_message_by_id(self.valid_message_2_by_other.message_id)
            is None
        )
        self.dialogue._update(self.valid_message_2_by_other)
        assert (
            self.dialogue.get_message_by_id(self.valid_message_2_by_other.message_id)
            == self.valid_message_2_by_other
        )

    def test_has_message_id(self):
        """Test the 'has_message_id' method."""
        assert self.dialogue._has_message_id(1) is False

        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue._has_message_id(1) is True

        assert self.dialogue._has_message_id(2) is False

    def test_update_positive(self):
        """Positive test for the 'update' method."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_outgoing_message == self.valid_message_1_by_self

    def test_update_positive_multiple_messages_by_self(self):
        """Positive test for the 'update' method: multiple messages by self is sent to the dialogue."""
        self.dialogue._update(self.valid_message_1_by_self)

        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        self.dialogue._update(valid_message_2_by_self)

        assert self.dialogue.last_message.message_id == 2

    def test_terminal_state_callback(self):
        """Test dialogue terminal state callback works."""
        called = False

        def callback(dialogue):
            nonlocal called
            called = True

        self.dialogue.add_terminal_state_callback(callback)
        self.dialogue._update(self.valid_message_1_by_self)

        self.dialogue.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )

        assert called

    def test_update_negative_is_valid_next_message_fails(self):
        """Negative test for the 'update' method: input message is invalid with respect to the dialogue."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=200,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(
            InvalidDialogueMessage,
            match=r"Message .* is invalid with respect to this dialogue. Error: Invalid message_id. Expected .*. Found 200.",
        ):
            self.dialogue._update(invalid_message_1_by_self)

        assert self.dialogue.last_outgoing_message is None

    def test_update_dialogue_negative_message_does_not_belong_to_dialogue(self):
        """Negative test for the 'update' method in dialogue with wrong message not belonging to dialogue."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(InvalidDialogueMessage) as cm:
            self.dialogue._update(invalid_message_1_by_self)
        assert str(cm.value) == (
            "The message 1 does not belong to this dialogue."
            "The dialogue reference of the message is {}, while the dialogue reference of the dialogue is {}".format(
                invalid_message_1_by_self.dialogue_reference,
                self.dialogue.dialogue_label.dialogue_reference,
            )
        )
        assert self.dialogue.is_empty

    def test_is_belonging_to_dialogue(self):
        """Test for the '_is_belonging_to_dialogue' method"""
        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        assert self.dialogue._is_belonging_to_dialogue(self.valid_message_1_by_self)
        assert not self.dialogue._is_belonging_to_dialogue(valid_message_2_by_self)

    def test_reply_positive(self):
        """Positive test for the 'reply' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        self.dialogue.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello Back",
        )

        assert self.dialogue.last_message.message_id == 2

    def test_reply_negative_empty_dialogue(self):
        """Negative test for the 'reply' method: target message is not in the dialogue."""
        with pytest.raises(ValueError) as cm:
            self.dialogue.reply(
                target_message=self.valid_message_1_by_self,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "Cannot reply in an empty dialogue!"
        assert self.dialogue.is_empty

    def test_reply_negative_target_does_not_exist(self):
        """Negative test for the 'reply' method: target is not in the dialogue."""
        self.dialogue._update(self.valid_message_1_by_self)
        with pytest.raises(ValueError) as cm:
            self.dialogue.reply(
                target=10,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "No target message found!"

    def test_reply_negative_target_message_target_mismatch(self):
        """Negative test for the 'reply' method: target message and target provided but do not match."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_message.message_id == 1

        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue.reply(
                target_message=self.valid_message_1_by_self,
                target=2,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "The provided target and target_message do not match."
        assert self.dialogue.last_message.message_id == 1

    def test_reply_negative_invalid_target(self):
        """Negative test for the 'reply' method: target message is not in the dialogue."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_message.message_id == 1

        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello There",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue.reply(
                target_message=invalid_message_1_by_self,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "The target message does not exist in this dialogue."
        assert self.dialogue.last_message.message_id == 1

    def test_is_valid_next_message_positive(self):
        """Positive test for the 'validate_next_message' method"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        result, msg = self.dialogue._validate_next_message(self.valid_message_3_by_self)
        assert result is True
        assert msg == "Message is valid with respect to this dialogue."

    def test_is_valid_next_message_negative_basic_validation_fails(self):
        """Negative test for the 'validate_next_message' method: basic_validation method fails"""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._validate_next_message(invalid_message_1_by_self)
        assert result is False
        assert msg == "Invalid message_id. Expected 1. Found 2."

    def test_is_valid_next_message_negative_additional_validation_fails(self):
        """Negative test for the 'validate_next_message' method: additional_validation method fails"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        invalid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=3,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        invalid_message_3_by_self.sender = self.agent_address
        invalid_message_3_by_self.to = self.opponent_address

        result, msg = self.dialogue._validate_next_message(invalid_message_3_by_self)
        assert result is False
        assert "Invalid target" in msg

    def test_is_valid_next_message_negative_is_valid_fails(self):
        """Negative test for the 'validate_next_message' method: is_valid method fails"""

        def failing_custom_validation(self, message: Message) -> Tuple[bool, str]:
            return False, "some reason"

        with patch.object(
            self.dialogue.__class__, "_custom_validation", failing_custom_validation
        ):
            result, msg = self.dialogue._validate_next_message(
                self.valid_message_1_by_self
            )

        assert result is False
        assert msg == "some reason"

    def test_basic_validation_positive(self):
        """Positive test for the '_basic_validation' method."""
        result, msg = self.dialogue._basic_validation(self.valid_message_1_by_self)
        assert result is True
        assert msg == "The initial message passes basic validation."

        self.dialogue._update(self.valid_message_1_by_self)

        result, msg = self.dialogue._basic_validation(self.valid_message_2_by_other)
        assert result is True
        assert msg == "The non-initial message passes basic validation."

    def test_basic_validation_negative_initial_message_invalid(self):
        """Negative test for the '_basic_validation' method: initial message is invalid."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        assert self.dialogue.is_empty
        result, msg = self.dialogue._basic_validation(invalid_message_1_by_self)
        assert result is False
        assert msg == "Invalid message_id. Expected 1. Found 2."

    @patch.object(BaseDialogue, "_validate_message_id", return_value=None)
    def test_basic_validation_negative_non_initial_message_invalid(self, *mocks):
        """Negative test for the '_basic_validation' method: non-initial message is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=-1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation(invalid_message_2_by_other)
        assert result is False
        assert msg == "Invalid target. Expected a non-zero integer. Found 0."

    def test_basic_validation_initial_message_positive(self):
        """Positive test for the '_basic_validation_initial_message' method."""
        result, msg = self.dialogue._basic_validation_initial_message(
            self.valid_message_1_by_self
        )
        assert result is True
        assert msg == "The initial message passes basic validation."

    def test_basic_validation_initial_message_negative_invalid_dialogue_reference(self):
        """Negative test for the '_basic_validation' method: input message has invalid dialogue reference."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert msg == "Invalid dialogue_reference[0]. Expected 1. Found 2."

    def test_basic_validation_initial_message_negative_invalid_message_id(self):
        """Negative test for the '_basic_validation' method: input message has invalid message id."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=200,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert re.match("Invalid message_id. Expected .*. Found 200.", msg)

    def test_basic_validation_initial_message_negative_invalid_target(self):
        """Negative test for the '_basic_validation_initial_message' method: input message has invalid target."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=1,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert msg == "Invalid target. Expected 0. Found 1."

    def test_basic_validation_initial_message_negative_invalid_performative(self):
        """Negative test for the '_basic_validation_initial_message' method: input message has invalid performative."""
        invalid_initial_msg = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.ERROR,
            error_code=DefaultMessage.ErrorCode.INVALID_MESSAGE,
            error_msg="some_error_message",
            error_data={"some_data": b"some_bytes"},
        )
        invalid_initial_msg.sender = self.agent_address
        invalid_initial_msg.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_initial_msg
        )
        assert result is False
        assert (
            msg
            == "Invalid initial performative. Expected one of {}. Found error.".format(
                self.dialogue.rules.initial_performatives
            )
        )

    def test_basic_validation_non_initial_message_positive(self):
        """Positive test for the '_basic_validation_non_initial_message' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        result, msg = self.dialogue._basic_validation_non_initial_message(
            self.valid_message_2_by_other
        )
        assert result is True
        assert msg == "The non-initial message passes basic validation."

    def test_basic_validation_non_initial_message_negative_invalid_dialogue_reference(
        self,
    ):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid dialogue reference."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid dialogue_reference[0]. Expected 1. Found 2."

    def test_basic_validation_non_initial_message_negative_invalid_message_id(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid message id."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=1000500000,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert re.match("Invalid message_id. Expected .*. Found 1000500000", msg)

    @patch.object(BaseDialogue, "_validate_message_id", return_value=None)
    def test_basic_validation_non_initial_message_negative_invalid_target_1(
        self, *mocks
    ):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has target less than 1."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid target. Expected a non-zero integer. Found 0."

    def test_basic_validation_non_initial_message_negative_invalid_target_2(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has target greater than the id of the last existing message."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=-1,
            target=2,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert "Invalid target. Expected a value less than " in msg

    @patch.object(BaseDialogue, "_validate_message_id", return_value=None)
    def test_basic_validation_non_initial_message_negative_invalid_performative(
        self, *mocks
    ):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid performative."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = StateUpdateMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=Mock(),
            target=1,
            performative=StateUpdateMessage.Performative.APPLY,
            amount_by_currency_id={},
            quantities_by_good_id={},
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid performative. Expected one of {}. Found {}.".format(
            self.dialogue.rules.get_valid_replies(
                self.valid_message_1_by_self.performative
            ),
            invalid_message_2_by_other.performative,
        )

    def test_update_dialogue_label_positive(self):
        """Positive test for the 'update_dialogue_label' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        new_label = DialogueLabel(
            (str(1), str(1)), self.valid_message_1_by_self.to, self.agent_address
        )
        self.dialogue._update_dialogue_label(new_label)

        assert self.dialogue.dialogue_label == new_label

    def test_update_dialogue_label_negative_invalid_existing_label(self):
        """Negative test for the 'update_dialogue_label' method: existing dialogue reference is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        new_label = DialogueLabel(
            (str(1), str(1)), self.valid_message_1_by_self.to, self.agent_address
        )
        self.dialogue._update_dialogue_label(new_label)
        assert self.dialogue.dialogue_label == new_label

        new_label = DialogueLabel(
            (str(1), str(2)), self.valid_message_1_by_self.to, self.agent_address
        )
        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue._update_dialogue_label(new_label)
        assert str(cm.value) == "Dialogue label cannot be updated."

        assert self.dialogue.dialogue_label != new_label

    def test_update_dialogue_label_negative_invalid_input_label(self):
        """Negative test for the 'update_dialogue_label' method: input dialogue label's dialogue reference is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)

        new_label = DialogueLabel(
            (str(2), ""), self.valid_message_1_by_self.to, self.agent_address
        )
        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue._update_dialogue_label(new_label)
        assert str(cm.value) == "Dialogue label cannot be updated."
        assert self.dialogue.dialogue_label != new_label

    def test___str__1(self):
        """Test the '__str__' method: dialogue is self initiated"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        self.dialogue._update(self.valid_message_3_by_self)

        dialogue_str = "Dialogue Label:\n1__agent 2_agent 1\nMessages:\nmessage_id=1, target=0, performative=bytes\nmessage_id=-1, target=1, performative=bytes\nmessage_id=2, target=-1, performative=bytes\n"

        assert str(self.dialogue) == dialogue_str

    def test___str__2(self):
        """Test the '__str__' method: dialogue is other initiated"""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        self.dialogue_opponent_started._update(valid_message_1_by_other)

        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=-1,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        self.dialogue_opponent_started._update(valid_message_2_by_self)

        valid_message_3_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=-1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        valid_message_3_by_other.sender = self.opponent_address
        valid_message_3_by_other.to = self.agent_address

        self.dialogue_opponent_started._update(valid_message_3_by_other)

        dialogue_str = "Dialogue Label:\n1_1_agent 2_agent 2\nMessages:\nmessage_id=1, target=0, performative=bytes\nmessage_id=-1, target=1, performative=bytes\nmessage_id=2, target=-1, performative=bytes\n"

        assert str(self.dialogue_opponent_started) == dialogue_str


class TestDialogueStats:
    """Test for DialogueStats."""

    def setup_method(self) -> None:
        """Initialise the environment to test DialogueStats."""
        self.agent_address = "agent 1"
        self.opponent_address = "agent 2"
        self.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.dialogue = Dialogue(dialogue_label=self.dialogue_label)
        end_states = frozenset(
            {Dialogue.EndState.SUCCESSFUL, Dialogue.EndState.FAILED}
        )  # type: FrozenSet[BaseDialogue.EndState]
        self.dialogue_stats = DialogueStats(end_states)

    def test_properties(self):
        """Test dialogue properties."""
        assert isinstance(self.dialogue_stats.self_initiated, dict)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

    def test_add_dialogue_endstate(self):
        """Test for the 'add_dialogue_endstate' method."""
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

        self.dialogue_stats.add_dialogue_endstate(Dialogue.EndState.SUCCESSFUL, True)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 1,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

        self.dialogue_stats.add_dialogue_endstate(Dialogue.EndState.FAILED, False)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 1,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 1,
        }


class TestDialoguesBase:
    """Test for Dialogues."""

    def setup_method(self):
        """Initialise the environment to test Dialogue."""
        self.agent_address = "agent 1"
        self.opponent_address = "agent 2"
        self.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.dialogue = Dialogue(dialogue_label=self.dialogue_label)
        self.own_dialogues = Dialogues(self.agent_address)
        self.opponent_dialogues = Dialogues(self.opponent_address)

        # convenient messages to reuse across tests
        self.valid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        self.valid_message_1_by_self.sender = self.agent_address
        self.valid_message_1_by_self.to = self.opponent_address

        self.valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        self.valid_message_2_by_other.sender = self.opponent_address
        self.valid_message_2_by_other.to = self.agent_address

    def test_dialogues_properties(self):
        """Test dialogue properties."""
        assert (
            self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label == dict()
        )
        assert self.own_dialogues.self_address == self.agent_address
        assert self.own_dialogues.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.own_dialogues.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.own_dialogues.message_class == DefaultMessage
        assert self.own_dialogues.dialogue_class == Dialogue

    def test_counterparty_from_message(self):
        """Test the 'counterparty_from_message' method."""
        assert (
            self.own_dialogues._counterparty_from_message(self.valid_message_1_by_self)
            == self.opponent_address
        )
        assert (
            self.own_dialogues._counterparty_from_message(self.valid_message_2_by_other)
            == self.opponent_address
        )

    def test_is_message_by_self(self):
        """Test the 'is_message_by_self' method."""
        assert self.own_dialogues._is_message_by_self(self.valid_message_1_by_self)
        assert not self.own_dialogues._is_message_by_self(self.valid_message_2_by_other)

    def test_is_message_by_other(self):
        """Test the 'is_message_by_other' method."""
        assert not self.own_dialogues._is_message_by_other(self.valid_message_1_by_self)
        assert self.own_dialogues._is_message_by_other(self.valid_message_2_by_other)

    def test_new_self_initiated_dialogue_reference(self):
        """Test the 'new_self_initiated_dialogue_reference' method."""
        self_initiated_ref = self.own_dialogues.new_self_initiated_dialogue_reference()
        assert (
            isinstance(self_initiated_ref[0], str)
            and self_initiated_ref[0] != ""
            and len(self_initiated_ref[0]) == DialogueLabel.NONCE_BYTES_NB * 2
        )
        assert self_initiated_ref[1] == ""
        self_initiated_ref_2 = (
            self.own_dialogues.new_self_initiated_dialogue_reference()
        )
        assert self_initiated_ref_2 != self_initiated_ref

    def test_create_positive(self):
        """Positive test for the 'create' method."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )
        self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )

    def test_create_negative_incorrect_performative_content_combination(self):
        """Negative test for the 'create' method: invalid performative and content combination (i.e. invalid message)."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )
        with pytest.raises(
            ValueError, match="Invalid initial performative. Expected one of"
        ):
            self.own_dialogues.create(
                self.opponent_address,
                DefaultMessage.Performative.ERROR,
                content=b"Hello",
            )
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_positive_new_dialogue_by_other(self):
        """Positive test for the 'update' method: the input message is for a new dialogue dialogue by other."""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        dialogue = self.own_dialogues.update(valid_message_1_by_other)

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )
        assert dialogue is not None
        assert dialogue.last_message.dialogue_reference == (str(1), "")
        assert dialogue.last_message.message_id == 1
        assert dialogue.last_message.target == 0
        assert dialogue.last_message.performative == DefaultMessage.Performative.BYTES
        assert dialogue.last_message.content == b"Hello"

    def test_update_positive_existing_dialogue(self):
        """Positive test for the 'update' method: the input message is for an existing dialogue."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        dialogue_reference = (
            msg.dialogue_reference[0],
            self.opponent_dialogues._generate_dialogue_nonce(),
        )
        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=dialogue_reference,
            message_id=dialogue.get_incoming_next_message_id(),
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )

        dialogue = self.own_dialogues.update(valid_message_2_by_other)

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )
        assert dialogue is not None
        assert dialogue.last_message.dialogue_reference == dialogue_reference
        assert dialogue.last_message.message_id == valid_message_2_by_other.message_id
        assert dialogue.last_message.target == valid_message_2_by_other.target
        assert dialogue.last_message.performative == DefaultMessage.Performative.BYTES
        assert dialogue.last_message.content == b"Hello back"

    def test_update_positive_existing_dialogue_2(self):
        """Positive test for the 'update' method: the input message is for an existing dialogue from the original sender."""
        msg_1, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        opponent_dialogue_1 = self.opponent_dialogues.update(msg_1)

        msg_2 = dialogue.reply(
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello again",
        )

        opponent_dialogue_2 = self.opponent_dialogues.update(msg_2)

        assert opponent_dialogue_1 == opponent_dialogue_2

    def test_update_negative_invalid_label(self):
        """Negative test for the 'update' method: dialogue is not extendable with the input message."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=0,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.sender = self.opponent_address
        invalid_message_1_by_other.to = self.agent_address

        assert not self.own_dialogues.update(invalid_message_1_by_other)

    def test_update_negative_new_dialogue_by_self(self):
        """Negative test for the 'update' method: the message is not by the counterparty."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(self.valid_message_1_by_self)
        assert (
            str(cm.value)
            == "Invalid 'update' usage. Update must only be used with a message by another agent."
        )

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_negative_no_to(self):
        """Negative test for the 'update' method: the 'to' field of the input message is not set."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.sender = self.opponent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert str(cm.value) == "The message's 'to' field is not set {}".format(
            invalid_message_1_by_other
        )

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_negative_no_sender(self):
        """Negative test for the 'update' method: the 'sender' field of the input message is not set."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.to = self.agent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert (
            str(cm.value)
            == "Invalid 'update' usage. Update must only be used with a message by another agent."
        )

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_negative_no_matching_to(self):
        """Negative test for the 'update' method: the 'to' field of the input message does not match self address."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.to = self.agent_address + "wrong_stuff"
        invalid_message_1_by_other.sender = self.opponent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert (
            str(cm.value)
            == "Message to and dialogue self address do not match. Got 'to=agent 1wrong_stuff' expected 'to=agent 1'."
        )

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_negative_invalid_message(self):
        """Negative test for the 'update' method: the message is invalid."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.ERROR,
            error_code=DefaultMessage.ErrorCode.INVALID_MESSAGE,
            error_msg="some_error_message",
            error_data={"some_data": b"some_bytes"},
        )
        invalid_message_1_by_other.sender = self.opponent_address
        invalid_message_1_by_other.to = self.agent_address

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        dialogue = self.own_dialogues.update(invalid_message_1_by_other)

        assert dialogue is None

        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_update_negative_existing_dialogue_non_nonexistent(self):
        """Negative test for the 'update' method: the dialogue referred by the input message does not exist."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        updated_dialogue = self.own_dialogues.update(invalid_message_2_by_other)

        assert updated_dialogue is None
        last_message = self.own_dialogues._dialogues_storage.get(
            dialogue.dialogue_label
        ).last_message
        assert (
            last_message.dialogue_reference[0] != ""
            and last_message.dialogue_reference[1] == ""
        )
        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).last_message.message_id
            == 1
        )
        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).last_message.target
            == 0
        )
        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).last_message.performative
            == DefaultMessage.Performative.BYTES
        )
        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).last_message.content
            == b"Hello"
        )

    def test_complete_dialogue_reference_positive(
        self,
    ):
        """Positive test for the '_complete_dialogue_reference' method."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)

        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).dialogue_label.dialogue_reference
            == valid_message_2_by_other.dialogue_reference
        )

    def test_complete_dialogue_reference_negative_incorrect_reference(
        self,
    ):
        """Negative test for the '_complete_dialogue_reference' method: the input message has invalid dialogue reference."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        wrong_own_ref = (
            "wrong reference"  # if correct, would be  msg.dialogue_reference[0]
        )
        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                wrong_own_ref,
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)
        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).dialogue_label.dialogue_reference
            == msg.dialogue_reference
        )

    def test_get_dialogue_positive_1(self):
        """Positive test for the 'get_dialogue' method: the dialogue is self initiated and the second message is by the other agent."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)

        assert (
            self.own_dialogues._dialogues_storage.get(
                dialogue.dialogue_label
            ).dialogue_label.dialogue_reference
            == valid_message_2_by_other.dialogue_reference
        )

        retrieved_dialogue = self.own_dialogues.get_dialogue(valid_message_2_by_other)

        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    def test_get_dialogue_positive_2(self):
        """Positive test for the 'get_dialogue' method: the dialogue is other initiated and the second message is by this agent."""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        dialogue = self.own_dialogues.update(valid_message_1_by_other)

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=dialogue.dialogue_label.dialogue_reference,
            message_id=valid_message_1_by_other.message_id + 1,
            target=valid_message_1_by_other.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.agent_address
        valid_message_2_by_other.to = self.opponent_address

        retrieved_dialogue = self.own_dialogues.get_dialogue(valid_message_2_by_other)

        assert retrieved_dialogue is not None
        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    @patch.object(BaseDialogue, "_validate_message_id", return_value=None)
    def test_get_dialogue_negative_invalid_reference(self, *mocks):
        """Negative test for the 'get_dialogue' method: the input message has invalid dialogue reference."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        dialogue = self.own_dialogues.update(valid_message_2_by_other)
        assert dialogue is not None

        invalid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_3_by_self.sender = self.agent_address
        invalid_message_3_by_self.to = self.opponent_address

        retrieved_dialogue = self.own_dialogues.get_dialogue(invalid_message_3_by_self)

        assert retrieved_dialogue is None

    def test_get_latest_label(self):
        """Positive test for the 'get_latest_label' method."""
        pass

    def test_get_dialogue_from_label_positive(self):
        """Positive test for the 'get_dialogue_from_label' method."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        retrieved_dialogue = self.own_dialogues.get_dialogue_from_label(
            dialogue.dialogue_label
        )
        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    def test_get_dialogue_from_label_negative_incorrect_input_label(self):
        """Negative test for the 'get_dialogue_from_label' method: the input dialogue label does not exist."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        incorrect_label = DialogueLabel(
            (str(1), "error"), self.opponent_address, self.agent_address
        )

        retrieved_dialogue = self.own_dialogues.get_dialogue_from_label(incorrect_label)
        assert retrieved_dialogue is None

    def test_create_self_initiated_positive(self):
        """Positive test for the '_create_self_initiated' method."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        self.own_dialogues._create_self_initiated(
            self.opponent_address, (str(1), ""), Dialogue.Role.ROLE1
        )
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )

    def test_create_self_initiated_negative_invalid_dialogue_reference(self):
        """Negative test for the '_create_self_initiated' method: invalid dialogue reference"""
        pass

    def test_create_opponent_initiated_positive(self):
        """Positive test for the '_create_opponent_initiated' method."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        self.own_dialogues._create_opponent_initiated(
            self.opponent_address, (str(1), ""), Dialogue.Role.ROLE2
        )
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 1
        )

    def test_create_opponent_initiated_negative_invalid_input_dialogue_reference(self):
        """Negative test for the '_create_opponent_initiated' method: input dialogue label has invalid dialogue reference."""
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

        try:
            self.own_dialogues._create_opponent_initiated(
                self.opponent_address, ("", str(1)), Dialogue.Role.ROLE2
            )
            result = True
        except AEAEnforceError:
            result = False

        assert not result
        assert (
            len(self.own_dialogues._dialogues_storage._dialogues_by_dialogue_label) == 0
        )

    def test_create_with_message(self):
        """Positive test for create with message."""
        msg = DefaultMessage(
            dialogue_reference=self.own_dialogues.new_self_initiated_dialogue_reference(),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        self.own_dialogues.create_with_message("opponent", msg)

    def test__create_positive(self):
        """Positive test for the '_create' method."""
        pass

    def test__create_negative_incomplete_dialogue_label_present(self):
        """Negative test for the '_create' method: incomplete dialogue label already present."""
        pass

    def test__create_negative_dialogue_label_present(self):
        """Negative test for the '_create' method: dialogue label already present."""
        pass

    def test_generate_dialogue_nonce(self):
        """Test the '_generate_dialogue_nonce' method."""
        nonce = self.own_dialogues._generate_dialogue_nonce()
        assert (
            isinstance(nonce, str)
            and nonce != ""
            and len(nonce) == DialogueLabel.NONCE_BYTES_NB * 2
        )
        second_nonce = self.own_dialogues._generate_dialogue_nonce()
        assert nonce != second_nonce

    def test_get_dialogues_with_counterparty(self):
        """Test get dialogues with counterparty."""
        assert (
            self.own_dialogues.get_dialogues_with_counterparty(self.opponent_address)
            == []
        )
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        assert self.own_dialogues.get_dialogues_with_counterparty(
            self.opponent_address
        ) == [dialogue]

    def test_setup(self):
        """Test dialogues.setup()."""
        self.own_dialogues.setup()

    def test_teardown(self):
        """Test dialogues.teardown()."""
        self.own_dialogues.teardown()


class TestPersistDialoguesStorage:
    """Test PersistDialoguesStorage."""

    def setup_method(self):
        """Initialise the environment to test PersistDialogueStorage."""
        self.agent_address = "agent 1"
        self.opponent_address = "agent 2"
        self.dialogues = Dialogues(
            self.agent_address, keep_terminal_state_dialogues=True
        )
        self.dialogues._dialogues_storage = PersistDialoguesStorage(self.dialogues)
        self.skill_component = Mock()
        self.skill_component.name = "test_component"
        self.skill_component.skill_id = PublicId("test", "test", "0.1.0")

        self.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.generic_storage = Storage("sqlite://:memory:", threaded=True)
        self.generic_storage.start()
        wait_for_condition(lambda: self.generic_storage.is_connected, timeout=10)
        self.skill_component.context.storage = self.generic_storage

    def teardown_method(self):
        """Tear down the environment to test PersistDialogueStorage."""
        self.generic_storage.stop()
        self.generic_storage.wait_completed(sync=True, timeout=10)

    def test_dialogue_serialize_deserialize(self):
        """Test dialogue dumped and restored."""
        msg = DefaultMessage(
            dialogue_reference=self.dialogues.new_self_initiated_dialogue_reference(),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        dialogue = self.dialogues.create_with_message("opponent", msg)
        data = dialogue.json()
        dialogue_restored = dialogue.__class__.from_json(dialogue.message_class, data)
        assert dialogue == dialogue_restored

    def test_dump_restore(self):
        """Test dump and load methods of the persists storage."""
        dialogues_storage = PersistDialoguesStorage(self.dialogues)
        dialogues_storage._skill_component = self.skill_component
        self.dialogues._dialogues_storage = dialogues_storage
        dialogues_storage._incomplete_to_complete_dialogue_labels[
            self.dialogue_label
        ] = self.dialogue_label
        self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        msg, dialogue = self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello2"
        )
        dialogue.reply(
            target_message=msg,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )
        assert dialogues_storage.dialogues_in_terminal_state
        assert dialogues_storage.dialogues_in_active_state
        assert dialogues_storage._dialogue_by_address
        assert dialogues_storage._incomplete_to_complete_dialogue_labels
        dialogues_storage.teardown()

        dialogues_storage_restored = PersistDialoguesStorage(self.dialogues)
        dialogues_storage_restored._skill_component = self.skill_component
        dialogues_storage_restored.setup()

        assert len(dialogues_storage._dialogue_by_address) == len(
            dialogues_storage_restored._dialogue_by_address
        )

        assert (
            dialogues_storage._incomplete_to_complete_dialogue_labels
            == dialogues_storage_restored._incomplete_to_complete_dialogue_labels
        )
        assert set(
            [str(i.dialogue_label) for i in dialogues_storage.dialogues_in_active_state]
        ) == set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage_restored.dialogues_in_active_state
            ]
        )
        assert set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage.dialogues_in_terminal_state
            ]
        ) == set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage_restored.dialogues_in_terminal_state
            ]
        )

        # test remove from storage on storeage.remove
        assert dialogues_storage_restored._terminal_dialogues_collection
        dialogue_label = dialogues_storage.dialogues_in_terminal_state[0].dialogue_label
        assert dialogues_storage_restored._terminal_dialogues_collection.get(
            str(dialogue_label)
        )
        dialogues_storage_restored.remove(dialogue_label)
        assert (
            dialogues_storage_restored._terminal_dialogues_collection.get(
                str(dialogue_label)
            )
            is None
        )

    def test_cleanup(self):
        """Test storage cleanup."""
        dialogues_storage = PersistDialoguesStorage(self.dialogues)
        dialogues_storage._skill_component = self.skill_component
        self.dialogues._dialogues_storage = dialogues_storage
        dialogues_storage._incomplete_to_complete_dialogue_labels[
            self.dialogue_label
        ] = self.dialogue_label
        self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        msg, dialogue = self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello2"
        )
        dialogue.reply(
            target_message=msg,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )
        assert dialogues_storage._dialogues_by_dialogue_label
        assert dialogues_storage._dialogue_by_address
        assert dialogues_storage._incomplete_to_complete_dialogue_labels
        assert dialogues_storage._terminal_state_dialogues_labels

        self.dialogues.cleanup()

        assert not dialogues_storage._dialogues_by_dialogue_label
        assert not dialogues_storage._dialogue_by_address
        assert not dialogues_storage._incomplete_to_complete_dialogue_labels
        assert not dialogues_storage._terminal_state_dialogues_labels


class TestPersistDialoguesStorageOffloading:
    """Test PersistDialoguesStorage."""

    def setup_method(self):
        """Initialise the environment to test PersistDialogueStorage."""
        self.agent_address = "agent 1"
        self.opponent_address = "agent 2"
        self.dialogues = Dialogues(
            self.agent_address, keep_terminal_state_dialogues=True
        )
        self.skill_component = Mock()
        self.skill_component.name = "test_component"
        self.skill_component.skill_id = PublicId("test", "test", "0.1.0")

        self.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.generic_storage = Storage("sqlite://:memory:", threaded=True)
        self.generic_storage.start()
        wait_for_condition(lambda: self.generic_storage.is_connected, timeout=10)
        self.skill_component.context.storage = self.generic_storage

    def teardown_method(self):
        """Tear down the environment to test PersistDialogueStorage."""
        self.generic_storage.stop()
        self.generic_storage.wait_completed(sync=True, timeout=10)
        self.dialogues.cleanup()

    def test_dump_restore(self):
        """Test dump and load methods of the persists storage."""
        dialogues_storage = PersistDialoguesStorageWithOffloading(self.dialogues)
        dialogues_storage._skill_component = self.skill_component
        self.dialogues._dialogues_storage = dialogues_storage
        dialogues_storage._incomplete_to_complete_dialogue_labels[
            self.dialogue_label
        ] = self.dialogue_label
        self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        msg, dialogue = self.dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello2"
        )
        dialogue.reply(
            target_message=msg,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )
        assert dialogues_storage.dialogues_in_terminal_state
        assert dialogues_storage.dialogues_in_active_state
        assert dialogues_storage._dialogue_by_address
        assert dialogues_storage._incomplete_to_complete_dialogue_labels
        dialogues_by_addr = dialogues_storage.get_dialogues_with_counterparty(
            dialogue.dialogue_label.dialogue_opponent_addr
        )
        dialogues_storage.teardown()

        dialogues_storage_restored = PersistDialoguesStorageWithOffloading(
            self.dialogues
        )
        dialogues_storage_restored._skill_component = self.skill_component
        dialogues_storage_restored.setup()

        assert len(dialogues_storage._dialogue_by_address) == len(
            dialogues_storage_restored._dialogue_by_address
        )

        assert (
            dialogues_storage._incomplete_to_complete_dialogue_labels
            == dialogues_storage_restored._incomplete_to_complete_dialogue_labels
        )
        assert set(
            [str(i.dialogue_label) for i in dialogues_storage.dialogues_in_active_state]
        ) == set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage_restored.dialogues_in_active_state
            ]
        )
        assert set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage.dialogues_in_terminal_state
            ]
        ) == set(
            [
                str(i.dialogue_label)
                for i in dialogues_storage_restored.dialogues_in_terminal_state
            ]
        )

        dialogue_label = dialogues_storage.dialogues_in_terminal_state[0].dialogue_label

        assert len(dialogues_by_addr) == len(
            dialogues_storage_restored.get_dialogues_with_counterparty(
                dialogue.dialogue_label.dialogue_opponent_addr
            )
        )
        # check get and cache
        assert not dialogues_storage_restored._terminal_state_dialogues_labels
        assert dialogues_storage_restored.get(dialogue_label) is not None
        assert dialogues_storage_restored._terminal_state_dialogues_labels

        # test remove from storage on storeage.remove
        assert dialogues_storage_restored._terminal_dialogues_collection
        assert (
            dialogues_storage_restored._terminal_dialogues_collection.get(
                str(dialogue_label)
            )
            is not None
        )
        dialogues_storage_restored.remove(dialogue_label)
        assert (
            dialogues_storage_restored._terminal_dialogues_collection.get(
                str(dialogue_label)
            )
            is None
        )
        assert dialogues_storage_restored.get(dialogue_label) is None


class TestBaseDialoguesStorage:
    """Test PersistDialoguesStorage."""

    def setup_method(self):
        """Initialise the environment to test Dialogue."""
        self.incomplete_reference = (str(1), "")
        self.complete_reference = (str(1), str(1))
        self.opponent_address = "agent 2"
        self.agent_address = "agent 1"

        self.dialogue_label = DialogueLabel(
            dialogue_reference=self.incomplete_reference,
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        self.dialogue = Dialogue(dialogue_label=self.dialogue_label)

        self.dialogue_label_opponent_started = DialogueLabel(
            dialogue_reference=self.complete_reference,
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.opponent_address,
        )
        self.dialogue_opponent_started = Dialogue(
            dialogue_label=self.dialogue_label_opponent_started
        )

        # convenient messages to reuse across tests
        self.valid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        self.valid_message_1_by_self.sender = self.agent_address
        self.valid_message_1_by_self.to = self.opponent_address

        self.storage = BasicDialoguesStorage(Mock())

    def test_dialogues_in_terminal_state_kept(self):
        """Test dialogues in terminal state handled properly."""
        assert not self.storage._incomplete_to_complete_dialogue_labels
        self.storage.add(self.dialogue)
        assert self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state
        assert len(self.storage._incomplete_to_complete_dialogue_labels) == 1

        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )

        assert not self.storage.dialogues_in_active_state
        assert self.storage.dialogues_in_terminal_state

        self.storage.remove(self.dialogue.dialogue_label)
        assert not self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._incomplete_to_complete_dialogue_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            len(
                self.storage._dialogue_by_address[
                    self.dialogue.dialogue_label.dialogue_opponent_addr
                ]
            )
            == 0
        )

    def test_dialogues_in_terminal_state_removed(self):
        """Test dialogues in terminal state handled properly."""
        self.storage._dialogues.is_keep_dialogues_in_terminal_state = False
        self.storage.add(self.dialogue)
        assert self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state

        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )

        assert not self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._incomplete_to_complete_dialogue_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            len(
                self.storage._dialogue_by_address[
                    self.dialogue.dialogue_label.dialogue_opponent_addr
                ]
            )
            == 0
        )

    def test_dialogues_complete_kept(self):
        """Test dialogues completed handled properly."""
        assert not self.storage._incomplete_to_complete_dialogue_labels
        self.storage.add(self.dialogue_opponent_started)
        assert self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state
        assert len(self.storage._incomplete_to_complete_dialogue_labels) == 1

        self.dialogue_opponent_started._update(self.valid_message_1_by_self)
        self.dialogue_opponent_started._incoming_messages.append(
            self.valid_message_1_by_self
        )
        self.dialogue_opponent_started.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.ERROR,
            error_code=ErrorCode.UNSUPPORTED_PROTOCOL,
            error_msg="oops",
            error_data={},
        )

        assert not self.storage.dialogues_in_active_state
        assert self.storage.dialogues_in_terminal_state

        self.storage.remove(self.dialogue_opponent_started.dialogue_label)
        assert not self.storage.dialogues_in_active_state
        assert not self.storage.dialogues_in_terminal_state
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._incomplete_to_complete_dialogue_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._terminal_state_dialogues_labels
        )
        assert (
            self.dialogue.dialogue_label
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            self.dialogue.dialogue_label.get_incomplete_version()
            not in self.storage._dialogues_by_dialogue_label
        )
        assert (
            len(
                self.storage._dialogue_by_address[
                    self.dialogue.dialogue_label.dialogue_opponent_addr
                ]
            )
            == 0
        )

    def teardown_method(self):
        """Tear down the environment to test BaseDialogueStorage."""


def test_find_caller_object():
    """Test find_caller_object."""

    class CustomSkillComponent(SkillComponent):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.storage = PersistDialoguesStorage(self)

        def setup(self):
            pass

        def teardown(self):
            pass

        @classmethod
        def parse_module(cls, *args, **kwargs):
            pass

    skill_component = CustomSkillComponent(Mock(), Mock(), Mock())
    assert skill_component.storage._skill_component == skill_component

    class CustomObject:
        def __init__(self, *args, **kwargs):
            self.component = find_caller_object(SkillComponent)

    custom_object = CustomObject()
    assert custom_object.component is None


def test_dialogues_keep_terminal_state_dialogues():
    """Test Dialogues keep_terminal_state_dialogues option."""
    initial = Dialogues._keep_terminal_state_dialogues
    dialogues = Dialogues(Mock(), keep_terminal_state_dialogues=True)
    assert dialogues.is_keep_dialogues_in_terminal_state is True
    assert Dialogues._keep_terminal_state_dialogues == initial

    dialogues = Dialogues(Mock(), keep_terminal_state_dialogues=False)
    assert dialogues.is_keep_dialogues_in_terminal_state is False
    assert Dialogues._keep_terminal_state_dialogues == initial



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Envelopes generation speed for Behaviour act test."""

from typing import Any, List, Tuple

import click
from aea_cli_benchmark.utils import (
    multi_run,
    number_of_runs_deco,
    output_format_deco,
    print_results,
    runtime_mode_deco,
    with_packages,
)

PACKAGES = [("protocol", "open_aea/signing"), ("protocol", "fetchai/default")]


@click.command(name="proactive")
@click.option(
    "--duration",
    default=3,
    help="Run time in seconds.",
    show_default=True,
)
@runtime_mode_deco
@number_of_runs_deco
@output_format_deco
def main(
    duration: int, runtime_mode: str, number_of_runs: int, output_format: str
) -> Any:
    """Check envelopes send/receive rate within behaviour."""
    with with_packages(PACKAGES):
        from aea_cli_benchmark.case_proactive.case import run

        parameters = {
            "Duration(seconds)": duration,
            "Runtime mode": runtime_mode,
            "Number of runs": number_of_runs,
        }

        def result_fn() -> List[Tuple[str, Any, Any, Any]]:
            return multi_run(
                int(number_of_runs),
                run,
                (duration, runtime_mode),
            )

        return print_results(output_format, parameters, result_fn)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains tests for CLI Registry add methods."""

import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from aea.cli.add import HashNotProvided
from aea.cli.registry.add import fetch_package
from aea.cli.registry.settings import REMOTE_IPFS
from aea.configurations.base import PublicId
from aea.helpers.base import cd
from aea.test_tools.test_cases import BaseAEATestCase


@mock.patch("aea.cli.registry.utils.request_api", return_value={"file": "url"})
@mock.patch("aea.cli.registry.add.download_file", return_value="filepath")
@mock.patch("aea.cli.registry.add.extract")
class FetchPackageTestCase(TestCase):
    """Test case for fetch_package method."""

    def test_fetch_package_positive(
        self, extract_mock, download_file_mock, request_api_mock
    ):
        """Test for fetch_package method positive result."""
        obj_type = "connection"
        public_id = PublicId.from_str("author/name:0.1.0")
        cwd = "cwd"
        dest_path = os.path.join("dest", "path", "package_folder_name")

        fetch_package(obj_type, public_id, cwd, dest_path)
        request_api_mock.assert_called_with(
            "GET", "/connections/author/name/0.1.0", params=None
        )
        download_file_mock.assert_called_once_with("url", "cwd")
        extract_mock.assert_called_once_with("filepath", os.path.join("dest", "path"))


class BaseTestAdd(BaseAEATestCase):
    """Test `aea add` command."""

    agent_name: str
    agent_dir: Path
    skill_path: Path
    skill_id: PublicId = PublicId(
        author="fetchai",
        name="echo",
        package_hash="bafybeia3ovoxmnipktwnyztie55itsuempnfeircw72jn62uojzry5pwsu",
    )

    @classmethod
    def setup_class(cls) -> None:
        """Setup test class."""

        super().setup_class()

        cls.agent_name = "agent"
        cls.create_agents(cls.agent_name)

        cls.agent_dir = cls.t / cls.agent_name
        cls.skill_path = (
            cls.packages_dir_path.absolute()
            / cls.skill_id.author
            / "skills"
            / cls.skill_id.name
        )


class TestAddFromHash(BaseTestAdd):
    """Test add using hash."""

    def test_from_hash(
        self,
    ) -> None:
        """Test run."""

        with cd(self.agent_dir), TemporaryDirectory() as temp_dir:
            download_path = Path(temp_dir, self.skill_id.name)
            shutil.copytree(self.skill_path, download_path)
            with mock.patch(
                "aea.cli.add.get_default_remote_registry", return_value=REMOTE_IPFS
            ), mock.patch(
                "aea.cli.add.fetch_ipfs", return_value=download_path
            ), mock.patch(
                "aea.cli.add._add_item_deps"
            ):
                result = self.run_cli_command(
                    "add",
                    "skill",
                    self.skill_id.hash,
                    "--remote",
                )

                assert result.exit_code == 0, result.output
                assert "Successfully added skill" in result.stdout

                assert (Path.cwd() / "vendor" / "fetchai" / "skills" / "echo").exists()


class TestAddFromHashIPFSFail(BaseTestAdd):
    """Test ipfs registry failing"""

    def test_hash_not_provided(
        self,
    ) -> None:
        """Test run."""

        with cd(self.agent_dir), TemporaryDirectory() as temp_dir:
            download_path = Path(temp_dir, self.skill_id.name)
            shutil.copytree(self.skill_path, download_path)
            with mock.patch(
                "aea.cli.add.get_default_remote_registry", return_value=REMOTE_IPFS
            ), mock.patch(
                "aea.cli.add.fetch_ipfs", side_effect=HashNotProvided
            ), mock.patch(
                "aea.cli.add.fetch_package", return_value=download_path
            ), mock.patch(
                "aea.cli.add._add_item_deps"
            ):
                result = self.run_cli_command(
                    "add",
                    "skill",
                    self.skill_id.hash,
                    "--remote",
                )

                assert result.exit_code == 0, result.output
                assert "Hash was not provided for" in result.stdout
                assert "Will try with http repository" in result.stdout
                assert "Successfully added skill" in result.stdout



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains AEA cli tests for P2PLibp2p connection."""

import json
import os
from pathlib import Path
from typing import List

from aea_ledger_cosmos import CosmosCrypto
from aea_ledger_ethereum.ethereum import EthereumCrypto as Ethereum

from aea.test_tools.test_cases import AEATestCaseEmpty

from packages.valory.connections import p2p_libp2p
from packages.valory.connections.p2p_libp2p.connection import (
    PUBLIC_ID as P2P_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.p2p_libp2p.consts import (
    LIBP2P_CERT_NOT_AFTER,
    LIBP2P_CERT_NOT_BEFORE,
)
from packages.valory.connections.p2p_libp2p.tests.base import (
    SKIP_WINDOWS,
    libp2p_log_on_failure_all,
)

p2p_libp2p_path = f"vendor.{p2p_libp2p.__name__.split('.', 1)[-1]}"
DEFAULT_NET_SIZE = 4
LIBP2P_LAUNCH_TIMEOUT = 20  # may downloads up to ~66Mb


@SKIP_WINDOWS
class BaseP2PLibp2pConnectionAEATest(AEATestCaseEmpty):
    """Base class for AEA CLI tests"""

    log_files: List[str] = []
    capture_log = True

    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent
    conn_key_file: str

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()
        cls.conn_key_file = os.path.join(os.path.abspath(os.getcwd()), "./conn_key.txt")
        cls.log_files = []

    @classmethod
    def teardown_class(cls) -> None:
        """Tear down the test"""
        cls.log_files.clear()
        super().teardown_class()

    def set_logging(self) -> None:
        """Set logging"""

        config_path = f"{p2p_libp2p_path}.config"
        log_file = f"libp2p_node_{self.agent_name}.log"
        log_file = os.path.join(os.path.abspath(os.getcwd()), log_file)
        self.set_config(f"{config_path}.log_file", log_file)
        self.log_files.append(log_file)

    def run_aea_cli_test(self) -> None:
        """Run the agent, check strings and terminate"""

        process = self.run_agent()
        is_running = self.is_running(process, timeout=LIBP2P_LAUNCH_TIMEOUT)
        assert is_running, "AEA not running within timeout!"

        check_strings = "Peer running in "
        missing_strings = self.missing_from_output(process, check_strings)
        assert not missing_strings


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionAEARunningEthereumConfigNode(
    BaseP2PLibp2pConnectionAEATest
):
    """Test AEA with p2p_libp2p connection is correctly run"""

    conn_key_file: str

    def test_agent(self) -> None:
        """Test with aea."""
        key_path = "ethereum_private_key.txt"
        self.generate_private_key(
            ledger_api_id=Ethereum.identifier, private_key_file=key_path
        )
        self.add_private_key(
            ledger_api_id=Ethereum.identifier, private_key_filepath=key_path
        )
        self.generate_private_key(
            CosmosCrypto.identifier, private_key_file=self.conn_key_file
        )
        self.add_private_key(
            CosmosCrypto.identifier,
            private_key_filepath=self.conn_key_file,
            connection=True,
        )
        self.add_item("connection", str(P2P_CONNECTION_PUBLIC_ID))
        self.run_cli_command("build", cwd=self._get_cwd())
        self.set_config("agent.default_ledger", Ethereum.identifier)
        self.nested_set_config(
            "agent.required_ledgers", [CosmosCrypto.identifier, Ethereum.identifier]
        )
        self.set_config("agent.default_connection", str(P2P_CONNECTION_PUBLIC_ID))

        setting_path = f"{p2p_libp2p_path}.cert_requests"
        settings = json.dumps(
            [
                {
                    "identifier": "acn",
                    "ledger_id": Ethereum.identifier,
                    "not_before": LIBP2P_CERT_NOT_BEFORE,
                    "not_after": LIBP2P_CERT_NOT_AFTER,
                    "public_key": CosmosCrypto.identifier,
                    "message_format": "{public_key}",
                    "save_path": ".certs/conn_cert.txt",
                }
            ]
        )
        self.set_config(setting_path, settings, type_="list")
        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

        self.set_logging()
        self.run_aea_cli_test()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for aea/components/base.py"""

import inspect
import itertools
import os
import re
import sys
from itertools import zip_longest
from pathlib import Path
from textwrap import dedent

import _strptime  # pylint: disbale=unsed-import # noqa: F401
import pytest

from aea.aea_builder import AEABuilder
from aea.components.base import Component, load_aea_package
from aea.configurations.base import (
    ConnectionConfig,
    PACKAGE_TYPE_TO_CONFIG_CLASS,
    ProtocolConfig,
)
from aea.configurations.data_types import PackageType, PublicId
from aea.configurations.loader import ConfigLoader
from aea.exceptions import AEAPackageLoadingError
from aea.helpers.base import cd
from aea.helpers.io import open_file
from aea.test_tools.test_cases import AEATestCase

from packages.valory.protocols.http.dialogues import (  # noqa: F401 # pylint: disable=unused-import
    HttpDialogue,
)

from tests.conftest import (
    CUR_PATH,
    ROOT_DIR,
    connection_config_files,
    contract_config_files,
    protocol_config_files,
    skill_config_files,
)


class TestComponentProperties:
    """Test accessibility of component properties."""

    def setup_class(self):
        """Setup test."""
        self.configuration = ProtocolConfig(
            "name", "author", "0.1.0", protocol_specification_id="some/author:0.1.0"
        )
        self.configuration.build_directory = "test"
        self.component = Component(configuration=self.configuration)
        self.directory = Path()
        self.component._directory = self.directory

    def test_component_type(self):
        """Test component type attribute."""
        assert self.component.component_type == self.configuration.component_type

    def test_is_vendor(self):
        """Test component type attribute."""
        assert self.component.is_vendor is False

    def test_prefix_import_path(self):
        """Test component type attribute."""
        assert self.component.prefix_import_path == "packages.author.protocols.name"

    def test_component_id(self):
        """Test component id."""
        assert self.component.component_id == self.configuration.component_id

    def test_public_id(self):
        """Test public id."""
        assert self.component.public_id == self.configuration.public_id

    def test_directory(self):
        """Test directory."""
        assert self.component.directory == self.directory

    def test_build_directory(self):
        """Test directory."""
        assert self.component.build_directory


def test_directory_setter():
    """Test directory."""
    configuration = ProtocolConfig(
        "author", "name", "0.1.0", protocol_specification_id="some/author:0.1.0"
    )
    component = Component(configuration=configuration)

    with pytest.raises(ValueError):
        component.directory

    new_path = Path("new_path")
    component.directory = new_path
    assert component.directory == new_path


def test_load_aea_package():
    """Test aea package load."""
    config = ConnectionConfig(
        "http_client", "valory", "0.5.0", protocols={PublicId("valory", "http")}
    )
    config.directory = (
        Path(ROOT_DIR) / "packages" / "valory" / "connections" / "http_client"
    )
    load_aea_package(config)


def test_load_aea_package_twice():
    """Test aea package load twice and ensure python objects stay the same."""
    config = ConnectionConfig(
        "http_client", "valory", "0.5.0", protocols={PublicId("valory", "http")}
    )
    config.directory = (
        Path(ROOT_DIR) / "packages" / "valory" / "connections" / "http_client"
    )
    # It doesn't matter if the package is already loaded.
    # We cannot safely remove it as references to other modules
    # would persist and get stale.
    if "packages.valory.connections.http_client.connection" not in sys.modules:
        load_aea_package(config)
        assert "packages.valory.connections.http_client.connection" not in sys.modules
        from packages.valory.connections.http_client.connection import (
            HTTPClientConnection,
        )

        assert "packages.valory.connections.http_client.connection" in sys.modules
        BaseHTTPCLientConnection = HTTPClientConnection
    else:
        members = inspect.getmembers(
            sys.modules["packages.valory.connections.http_client.connection"],
            inspect.isclass,
        )
        BaseHTTPCLientConnection = [
            pairs[1] for pairs in members if pairs[0] == "HTTPClientConnection"
        ][0]
    # second time
    load_aea_package(config)
    from packages.valory.connections.http_client.connection import HTTPClientConnection

    assert BaseHTTPCLientConnection is HTTPClientConnection


class TestLoadingErrorWithUndeclaredDependency(AEATestCase):
    """Test that an error is raised when not declared dependencies are detected in import statements of modules."""

    path_to_aea = Path(CUR_PATH) / "data" / "dummy_aea"

    matching_content = re.escape(
        dedent(
            f"""        aea.exceptions.AEAPackageLoadingError: found the following packages that are imported by some module but not declared as dependencies of package (skill, dummy_author/dummy:0.1.0):

        - connection some_author/some_connection used in:
            - dummy_subpackage{os.path.sep}__init__.py
        - protocol some_author/some_protocol used in:
            - dummy_subpackage{os.path.sep}__init__.py
        """
        )
    )

    def _add_undeclared_dependency_in_dummy_skill_imports(self):
        """
        Set up the dummy skill for the test.

        Add package import statements, in some module of the dummy skill, of packages
        that are not declared as dependencies of the dummy skill.
        """
        cwd = self._get_cwd()
        module_path = Path(cwd, "skills", "dummy", "dummy_subpackage", "__init__.py")
        module_content = module_path.read_text()
        module_content += "\nimport packages.some_author.connections.some_connection"
        module_content += "\nfrom packages.some_author.protocols.some_protocol"
        module_path.write_text(module_content)
        self.run_cli_command("fingerprint", "by-path", "skills/dummy", cwd=cwd)

    def test_run(self):
        """Run the test."""
        self._add_undeclared_dependency_in_dummy_skill_imports()
        with cd(self._get_cwd()):
            builder = AEABuilder.from_aea_project(Path(self._get_cwd()))
            with pytest.raises(AEAPackageLoadingError, match=self.matching_content):
                builder.build()


class TestLoadingErrorWithUnusedDependency(AEATestCase):
    """Test that an error is raised when not used dependencies are detected in import statements of modules."""

    path_to_aea = Path(CUR_PATH) / "data" / "dummy_aea"

    matching_content = re.escape(
        dedent(
            """        aea.exceptions.AEAPackageLoadingError: The following dependencies are not used in any Python module of the package (skill, dummy_author/dummy:0.1.0):
        - connection fetchai/local
        - protocol fetchai/fipa
"""
        )
    )

    def _add_unused_dependency_in_dummy_skill_imports(self):
        """
        Set up the dummy skill for the test.

        Add a dummy dependency that is not used in the skill.
        """
        cwd = self._get_cwd()
        configuration_path = Path(cwd, "skills", "dummy", "skill.yaml")
        configuration_content = configuration_path.read_text()
        configuration_content = configuration_content.replace(
            "protocols:", "protocols:\n- fetchai/fipa:1.0.0"
        )
        configuration_content = configuration_content.replace(
            "connections: []", "connections:\n- fetchai/local:0.20.0"
        )
        configuration_path.write_text(configuration_content)
        self.run_cli_command("fingerprint", "by-path", "skills/dummy", cwd=cwd)

    def test_run(self):
        """Run the test."""
        self._add_unused_dependency_in_dummy_skill_imports()
        with cd(self._get_cwd()):
            builder = AEABuilder.from_aea_project(Path(self._get_cwd()))
            with pytest.raises(AEAPackageLoadingError, match=self.matching_content):
                builder.build()


@pytest.mark.parametrize(
    "component_type,config_file_path",
    itertools.chain.from_iterable(
        [
            zip_longest([], files, fillvalue=component_type)
            for files, component_type in [
                (protocol_config_files, PackageType.PROTOCOL),
                (contract_config_files, PackageType.CONTRACT),
                (connection_config_files, PackageType.CONNECTION),
                (skill_config_files, PackageType.SKILL),
            ]
        ]
    ),
)
def test_load_all_aea_protocol_packages(
    component_type: PackageType, config_file_path: str
) -> None:
    """Load all AEA component packages."""

    to_be_skipped = {os.path.join(CUR_PATH, "data", "local-connection.yaml")}

    if config_file_path in to_be_skipped:
        pytest.skip(
            f"test not supported for configurations outside package: {config_file_path}"
        )

    configuration_loader = ConfigLoader.from_configuration_type(
        component_type, PACKAGE_TYPE_TO_CONFIG_CLASS
    )
    with open_file(config_file_path) as fp:
        configuration_object = configuration_loader.load(fp)
        directory = Path(config_file_path).parent
        configuration_object.directory = directory
        load_aea_package(configuration_object)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the Chainlist RPC enrichment module."""

import json
from unittest.mock import MagicMock, patch

import pytest
from aea_ledger_ethereum.chainlist import (
    ChainlistRPC,
    MAX_CHAINLIST_CANDIDATES,
    RPCNode,
    _filter_candidates,
    _is_template_url,
    _normalize_url,
    _rank_and_select,
    enrich_rpc_urls,
    probe_rpc,
)

# ---------------------------------------------------------------------------
# URL helpers
# ---------------------------------------------------------------------------


class TestNormalizeUrl:
    """Tests for _normalize_url."""

    def test_strips_trailing_slash(self) -> None:
        """Trailing slash is stripped."""
        assert _normalize_url("https://rpc.example.com/") == "https://rpc.example.com"

    def test_lowercases(self) -> None:
        """URL is lowercased."""
        assert _normalize_url("https://RPC.Example.COM") == "https://rpc.example.com"

    def test_no_change_needed(self) -> None:
        """Already-normalized URL is unchanged."""
        assert _normalize_url("https://rpc.example.com") == "https://rpc.example.com"


class TestIsTemplateUrl:
    """Tests for _is_template_url."""

    def test_dollar_brace(self) -> None:
        """Dollar-brace template is detected."""
        assert _is_template_url("https://rpc.example.com/${API_KEY}") is True

    def test_plain_brace(self) -> None:
        """Plain-brace template is detected."""
        assert _is_template_url("https://rpc.example.com/{key}") is True

    def test_no_template(self) -> None:
        """Plain URL is not a template."""
        assert _is_template_url("https://rpc.example.com") is False


# ---------------------------------------------------------------------------
# probe_rpc
# ---------------------------------------------------------------------------


class TestProbeRpc:
    """Tests for probe_rpc."""

    @patch("aea_ledger_ethereum.chainlist.time.monotonic", side_effect=[1.0, 1.5])
    @patch("aea_ledger_ethereum.chainlist.urllib.request.urlopen")
    def test_success(self, mock_urlopen: MagicMock, _mono: MagicMock) -> None:
        """Successful probe returns (url, latency, block)."""
        mock_resp = MagicMock()
        mock_resp.read.return_value = json.dumps(
            {"jsonrpc": "2.0", "result": "0x1A4B5C", "id": 1}
        ).encode()
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = mock_resp

        result = probe_rpc("https://rpc.example.com")
        assert result is not None
        url, latency, block = result
        assert url == "https://rpc.example.com"
        assert latency == 500.0
        assert block == 0x1A4B5C

    @patch("aea_ledger_ethereum.chainlist.urllib.request.urlopen")
    def test_timeout_returns_none(self, mock_urlopen: MagicMock) -> None:
        """Timeout returns None."""
        mock_urlopen.side_effect = TimeoutError("timed out")
        assert probe_rpc("https://slow.example.com") is None

    @patch("aea_ledger_ethereum.chainlist.urllib.request.urlopen")
    def test_zero_block_returns_none(self, mock_urlopen: MagicMock) -> None:
        """Block number 0x0 returns None."""
        mock_resp = MagicMock()
        mock_resp.read.return_value = json.dumps(
            {"jsonrpc": "2.0", "result": "0x0", "id": 1}
        ).encode()
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = mock_resp
        assert probe_rpc("https://rpc.example.com") is None

    @patch("aea_ledger_ethereum.chainlist.urllib.request.urlopen")
    def test_null_result_returns_none(self, mock_urlopen: MagicMock) -> None:
        """Null result returns None."""
        mock_resp = MagicMock()
        mock_resp.read.return_value = json.dumps(
            {"jsonrpc": "2.0", "result": None, "id": 1}
        ).encode()
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = mock_resp
        assert probe_rpc("https://rpc.example.com") is None


# ---------------------------------------------------------------------------
# RPCNode
# ---------------------------------------------------------------------------


class TestRPCNode:
    """Tests for RPCNode dataclass."""

    def test_tracking_limited(self) -> None:
        """Tracking 'limited' is flagged."""
        node = RPCNode(url="https://x.com", is_working=True, tracking="limited")
        assert node.is_tracking is True

    def test_tracking_yes(self) -> None:
        """Tracking 'yes' is flagged."""
        node = RPCNode(url="https://x.com", is_working=True, tracking="yes")
        assert node.is_tracking is True

    def test_tracking_none(self) -> None:
        """Tracking 'none' is not flagged."""
        node = RPCNode(url="https://x.com", is_working=True, tracking="none")
        assert node.is_tracking is False

    def test_tracking_default(self) -> None:
        """Default tracking (None) is not flagged."""
        node = RPCNode(url="https://x.com", is_working=True)
        assert node.is_tracking is False


# ---------------------------------------------------------------------------
# _filter_candidates
# ---------------------------------------------------------------------------


class TestFilterCandidates:
    """Tests for _filter_candidates."""

    def test_filters_non_https(self) -> None:
        """Non-HTTPS URLs are filtered out."""
        nodes = [
            RPCNode(url="http://insecure.com", is_working=True),
            RPCNode(url="wss://ws.com", is_working=True),
            RPCNode(url="https://good.com", is_working=True),
        ]
        result = _filter_candidates(nodes, set())
        assert result == ["https://good.com"]

    def test_filters_templates(self) -> None:
        """Template URLs are filtered out."""
        nodes = [
            RPCNode(url="https://rpc.com/${API_KEY}", is_working=True),
            RPCNode(url="https://good.com", is_working=True),
        ]
        result = _filter_candidates(nodes, set())
        assert result == ["https://good.com"]

    def test_deduplicates_existing(self) -> None:
        """Existing URLs are deduplicated."""
        nodes = [
            RPCNode(url="https://already.com", is_working=True),
            RPCNode(url="https://new.com", is_working=True),
        ]
        existing = {"https://already.com"}
        result = _filter_candidates(nodes, existing)
        assert result == ["https://new.com"]

    def test_max_candidates_limit(self) -> None:
        """Candidate list is capped at MAX_CHAINLIST_CANDIDATES."""
        nodes = [
            RPCNode(url=f"https://rpc{i}.com", is_working=True)
            for i in range(MAX_CHAINLIST_CANDIDATES + 10)
        ]
        result = _filter_candidates(nodes, set())
        assert len(result) == MAX_CHAINLIST_CANDIDATES


# ---------------------------------------------------------------------------
# _rank_and_select
# ---------------------------------------------------------------------------


class TestRankAndSelect:
    """Tests for _rank_and_select."""

    def test_sorts_by_latency(self) -> None:
        """Results are sorted by latency (fastest first)."""
        results = [
            ("https://slow.com", 200.0, 1000),
            ("https://fast.com", 10.0, 1000),
            ("https://med.com", 80.0, 1000),
        ]
        selected = _rank_and_select(results, [], chain_id=100, max_results=3)
        assert selected == ["https://fast.com", "https://med.com", "https://slow.com"]

    def test_filters_stale(self) -> None:
        """Stale RPCs lagging behind median block are filtered out."""
        results = [
            ("https://fresh.com", 50.0, 1000),
            ("https://stale.com", 30.0, 900),  # 100 blocks behind
            ("https://ok.com", 60.0, 999),
        ]
        selected = _rank_and_select(results, [], chain_id=100, max_results=5)
        assert selected == ["https://fresh.com", "https://ok.com"]

    def test_respects_max_results(self) -> None:
        """Output is capped at max_results."""
        results = [(f"https://rpc{i}.com", float(i), 1000) for i in range(10)]
        selected = _rank_and_select(results, [], chain_id=100, max_results=3)
        assert len(selected) == 3


# ---------------------------------------------------------------------------
# ChainlistRPC
# ---------------------------------------------------------------------------


class TestChainlistRPC:
    """Tests for ChainlistRPC."""

    def test_get_rpcs_parses_nodes(self) -> None:
        """get_rpcs parses RPC entries into RPCNode objects."""
        with patch.object(ChainlistRPC, "fetch_data"):
            cl = ChainlistRPC()
            cl._data = [
                {
                    "chainId": 100,
                    "rpc": [
                        {"url": "https://rpc1.example.com", "tracking": "none"},
                        {"url": "https://rpc2.example.com", "tracking": "yes"},
                    ],
                }
            ]
            result = cl.get_rpcs(100)
            assert len(result) == 2
            assert result[0].url == "https://rpc1.example.com"
            assert result[1].tracking == "yes"

    def test_get_rpcs_chain_not_found(self) -> None:
        """Unknown chain_id returns empty list."""
        with patch.object(ChainlistRPC, "fetch_data"):
            cl = ChainlistRPC()
            cl._data = [{"chainId": 1, "rpc": []}]
            assert cl.get_rpcs(999) == []

    @patch.object(ChainlistRPC, "get_rpcs")
    @patch("aea_ledger_ethereum.chainlist.probe_rpc")
    def test_get_validated_rpcs_full_pipeline(
        self, mock_probe: MagicMock, mock_get_rpcs: MagicMock
    ) -> None:
        """Full pipeline: filter → probe → rank → select."""
        mock_get_rpcs.return_value = [
            RPCNode(url="https://rpc.com/${API_KEY}", is_working=True),  # filtered
            RPCNode(url="https://good1.com", is_working=True),
            RPCNode(url="https://good2.com", is_working=True),
        ]
        mock_probe.side_effect = [
            ("https://good1.com", 100.0, 1000),
            ("https://good2.com", 50.0, 1000),
        ]

        cl = ChainlistRPC()
        result = cl.get_validated_rpcs(100, existing_rpcs=[])
        assert result == ["https://good2.com", "https://good1.com"]

    @patch.object(ChainlistRPC, "get_rpcs")
    def test_returns_empty_on_no_rpcs(self, mock_get_rpcs: MagicMock) -> None:
        """Returns empty when no RPCs exist for chain."""
        mock_get_rpcs.return_value = []
        cl = ChainlistRPC()
        assert cl.get_validated_rpcs(100, existing_rpcs=[]) == []

    @patch.object(ChainlistRPC, "get_rpcs")
    @patch("aea_ledger_ethereum.chainlist.probe_rpc")
    def test_returns_empty_when_all_probes_fail(
        self, mock_probe: MagicMock, mock_get_rpcs: MagicMock
    ) -> None:
        """Returns empty when all probes fail."""
        mock_get_rpcs.return_value = [
            RPCNode(url="https://dead.com", is_working=True),
        ]
        mock_probe.return_value = None
        cl = ChainlistRPC()
        assert cl.get_validated_rpcs(100, existing_rpcs=[]) == []


# ---------------------------------------------------------------------------
# Fetch data and caching
# ---------------------------------------------------------------------------


class TestFetchData:
    """Tests for fetch_data caching behavior."""

    def test_uses_cache_when_valid(self, tmp_path: "pytest.TempPathFactory") -> None:
        """Valid cache is used without network fetch."""
        cache_file = tmp_path / "rpcs.json"
        cache_data = [{"chainId": 100, "name": "Test", "rpc": []}]
        cache_file.write_text(json.dumps(cache_data))

        with patch("aea_ledger_ethereum.chainlist._CACHE_PATH", cache_file):
            with patch("aea_ledger_ethereum.chainlist.urllib.request.urlopen") as m:
                cl = ChainlistRPC()
                cl.fetch_data()
                m.assert_not_called()
                assert cl._data == cache_data

    def test_fetches_when_cache_expired(
        self, tmp_path: "pytest.TempPathFactory"
    ) -> None:
        """Expired cache triggers a fresh network fetch."""
        cache_file = tmp_path / "rpcs.json"
        cache_dir = tmp_path
        fresh_data = [{"chainId": 100, "name": "Fresh", "rpc": []}]

        mock_resp = MagicMock()
        mock_resp.read.return_value = json.dumps(fresh_data).encode()
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)

        with patch("aea_ledger_ethereum.chainlist._CACHE_PATH", cache_file), patch(
            "aea_ledger_ethereum.chainlist._CACHE_DIR", cache_dir
        ), patch(
            "aea_ledger_ethereum.chainlist.urllib.request.urlopen",
            return_value=mock_resp,
        ):
            cl = ChainlistRPC()
            cl.fetch_data(force_refresh=True)
            assert cl._data == fresh_data

    def test_falls_back_to_expired_cache(
        self, tmp_path: "pytest.TempPathFactory"
    ) -> None:
        """Network failure falls back to expired cache."""
        cache_file = tmp_path / "rpcs.json"
        old_data = [{"chainId": 100, "name": "Old", "rpc": []}]
        cache_file.write_text(json.dumps(old_data))

        with patch("aea_ledger_ethereum.chainlist._CACHE_PATH", cache_file), patch(
            "aea_ledger_ethereum.chainlist.CACHE_TTL", 0
        ), patch(
            "aea_ledger_ethereum.chainlist.urllib.request.urlopen",
            side_effect=Exception("network error"),
        ):
            cl = ChainlistRPC()
            cl.fetch_data()
            assert cl._data == old_data


# ---------------------------------------------------------------------------
# enrich_rpc_urls (main entry point)
# ---------------------------------------------------------------------------


class TestEnrichRpcUrls:
    """Tests for enrich_rpc_urls."""

    def test_no_chain_id_returns_unchanged(self) -> None:
        """No chain_id returns URLs unchanged."""
        urls = ["https://rpc1.com"]
        assert enrich_rpc_urls(urls) is urls

    def test_already_at_max_returns_unchanged(self) -> None:
        """Already at max RPCs returns URLs unchanged."""
        urls = [f"https://rpc{i}.com" for i in range(20)]
        assert enrich_rpc_urls(urls, chain_id=100) is urls

    @patch("aea_ledger_ethereum.chainlist.ChainlistRPC")
    def test_enriches_with_chainlist_rpcs(self, mock_cl_cls: MagicMock) -> None:
        """Chainlist RPCs are appended to the original list."""
        mock_cl = mock_cl_cls.return_value
        mock_cl.get_validated_rpcs.return_value = [
            "https://extra1.com",
            "https://extra2.com",
        ]

        result = enrich_rpc_urls(["https://original.com"], chain_id=100)
        assert result == [
            "https://original.com",
            "https://extra1.com",
            "https://extra2.com",
        ]

    @patch("aea_ledger_ethereum.chainlist.ChainlistRPC")
    def test_survives_failure(self, mock_cl_cls: MagicMock) -> None:
        """Enrichment failure returns original URLs."""
        mock_cl_cls.side_effect = Exception("network error")
        urls = ["https://rpc1.com"]
        result = enrich_rpc_urls(urls, chain_id=100)
        assert result == ["https://rpc1.com"]

    @patch("aea_ledger_ethereum.chainlist.ChainlistRPC")
    def test_no_extra_rpcs_returns_original(self, mock_cl_cls: MagicMock) -> None:
        """No extra RPCs found returns original URLs."""
        mock_cl = mock_cl_cls.return_value
        mock_cl.get_validated_rpcs.return_value = []

        urls = ["https://rpc1.com"]
        result = enrich_rpc_urls(urls, chain_id=100)
        assert result == ["https://rpc1.com"]


# ---------------------------------------------------------------------------
# Integration: RPCRotationMiddleware.build with chain_id → Chainlist enrichment
# ---------------------------------------------------------------------------


class TestRotationMixinChainlistIntegration:
    """Test that RPCRotationMiddleware enriches RPCs when chain_id is given."""

    @patch("aea_ledger_ethereum.chainlist.ChainlistRPC")
    def test_init_rotation_enriches_with_chain_id(self, mock_cl_cls: MagicMock) -> None:
        """build() enriches RPCs when chain_id is provided."""
        from unittest.mock import MagicMock as _MM

        from aea_ledger_ethereum.rpc_rotation import RPCRotationMiddleware

        mock_cl = mock_cl_cls.return_value
        mock_cl.get_validated_rpcs.return_value = ["https://extra.com"]

        mock_w3 = _MM()
        mw = RPCRotationMiddleware.build(
            mock_w3,
            rpc_urls=["https://original.com"],
            request_kwargs={},
            chain_id=100,
        )

        assert mw.rpc_count == 2
        assert mw._rpc_urls == ["https://original.com", "https://extra.com"]

    @patch("aea_ledger_ethereum.chainlist.ChainlistRPC")
    def test_init_rotation_no_chain_id_no_enrichment(
        self, mock_cl_cls: MagicMock
    ) -> None:
        """build() skips enrichment when no chain_id is given."""
        from unittest.mock import MagicMock as _MM

        from aea_ledger_ethereum.rpc_rotation import RPCRotationMiddleware

        mock_w3 = _MM()
        mw = RPCRotationMiddleware.build(
            mock_w3,
            rpc_urls=["https://only.com"],
            request_kwargs={},
        )

        assert mw.rpc_count == 1
        mock_cl_cls.assert_not_called()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check amount of time for acn connection start."""

import asyncio
import logging
import os
import time
from statistics import mean
from tempfile import TemporaryDirectory
from typing import Callable, List, Tuple, Union

from aea_cli_benchmark.case_acn_startup.utils import (
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    _make_libp2p_mailbox_connection,
)

from aea.connections.base import Connection


async def _multi_run_connect(con_maker: Callable[..., Connection], times: int) -> float:
    timings = []
    logging.basicConfig(level=logging.CRITICAL)
    for _ in range(times):
        con = con_maker()
        start_time = time.time()
        await con.connect()
        assert con.is_connected
        timings.append(time.time() - start_time)
        await con.disconnect()
    return mean(timings)


async def _run_libp2p_node(times: int) -> float:
    return await _multi_run_connect(
        con_maker=lambda: _make_libp2p_connection("."), times=times
    )


async def _run_p2p_client(times: int) -> float:
    node_con = _make_libp2p_connection(".", delegate=True)
    await node_con.connect()
    try:
        return await _multi_run_connect(
            con_maker=lambda: _make_libp2p_client_connection(
                peer_public_key=node_con.node.pub, data_dir="."
            ),
            times=times,
        )
    finally:
        await node_con.disconnect()


async def _run_p2p_mailbox(times: int) -> float:
    node_con = _make_libp2p_connection(".", mailbox=True)
    await node_con.connect()
    try:
        return await _multi_run_connect(
            con_maker=lambda: _make_libp2p_mailbox_connection(
                peer_public_key=node_con.node.pub, data_dir="."
            ),
            times=times,
        )
    finally:
        await node_con.disconnect()


def run(connection: str, run_times: int = 10) -> List[Tuple[str, Union[int, float]]]:
    """Check construction time and memory usage."""
    cwd = os.getcwd()
    try:
        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            if connection == "p2pnode":
                coro = _run_libp2p_node(run_times)
            elif connection == "client":
                coro = _run_p2p_client(run_times)
            elif connection == "mailbox":
                coro = _run_p2p_mailbox(run_times)
            else:
                raise ValueError(f"Unsupported connection: {connection}")

            connect_time = asyncio.get_event_loop().run_until_complete(coro)

            return [
                ("connect time (seconds)", connect_time),
            ]
    finally:
        os.chdir(cwd)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea run' subcommand."""

from contextlib import contextmanager, nullcontext
from pathlib import Path
from typing import Generator, List, Optional, Sequence, Tuple, cast

import click

from aea import __version__
from aea.aea import AEA
from aea.aea_builder import AEABuilder, DEFAULT_ENV_DOTFILE
from aea.cli.install import do_install
from aea.cli.utils.click_utils import ConnectionsOption, password_option
from aea.cli.utils.constants import AEA_LOGO, REQUIREMENTS
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project
from aea.cli.utils.exceptions import aev_flag_depreaction
from aea.cli.utils.package_utils import list_available_packages
from aea.configurations.base import ComponentType, PublicId
from aea.configurations.manager import AgentConfigManager
from aea.connections.base import Connection
from aea.contracts.base import Contract
from aea.exceptions import AEAWalletNoAddressException
from aea.helpers.base import load_env_file
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.helpers.profiling import Profiling
from aea.protocols.base import Message, Protocol
from aea.protocols.dialogue.base import Dialogue, DialogueLabel
from aea.skills.base import Behaviour, Handler, Model, Skill


@click.command()
@password_option()
@click.option(
    "--connections",
    "connection_ids",
    cls=ConnectionsOption,
    required=False,
    default=None,
    help="The connection names to use for running the agent. Must be declared in the agent's configuration file.",
)
@click.option(
    "--env",
    "env_file",
    type=click.Path(),
    required=False,
    default=DEFAULT_ENV_DOTFILE,
    help="Specify an environment file (default: .env)",
)
@click.option(
    "--install-deps",
    "is_install_deps",
    is_flag=True,
    required=False,
    default=False,
    help="Install all the dependencies before running the agent.",
)
@click.option(
    "--profiling",
    "profiling",
    required=False,
    default=0,
    help="Enable profiling, print profiling every amount of seconds",
)
@click.option(
    "--memray",
    "memray_flag",
    is_flag=True,
    required=False,
    default=False,
    help="Enable memray tracing, create a bin file with the memory dump",
)
@click.option(
    "--exclude-connections",
    "exclude_connection_ids",
    cls=ConnectionsOption,
    required=False,
    default=None,
    help="The connection names to disable while running the agent. Must be declared in the agent's configuration file.",
)
@click.option(
    "--aev",
    "apply_environment_variables",
    required=False,
    is_flag=True,
    default=False,
    help="Populate Agent configs from Environment variables.",
)
@click.pass_context
@check_aea_project
def run(  # pylint: disable=too-many-positional-arguments
    click_context: click.Context,
    connection_ids: List[PublicId],
    exclude_connection_ids: List[PublicId],
    env_file: str,
    is_install_deps: bool,
    apply_environment_variables: bool,
    profiling: int,
    memray_flag: bool,
    password: str,
) -> None:
    """Run the agent."""
    if apply_environment_variables:
        aev_flag_depreaction()

    if connection_ids and exclude_connection_ids:
        raise click.ClickException(
            "Please use only one of --connections or --exclude-connections, not both!"
        )

    ctx = cast(Context, click_context.obj)
    profiling = int(profiling)
    if exclude_connection_ids:
        connection_ids = _calculate_connection_ids(ctx, exclude_connection_ids)

    profiling_context = (
        _profiling_context(period=profiling) if profiling > 0 else nullcontext()
    )

    memray_context = nullcontext()
    if memray_flag:
        try:
            import memray  # type: ignore # pylint: disable=import-error,import-outside-toplevel

            memray_context = memray.Tracker("memray_profiling.bin")
        except ModuleNotFoundError:
            click.echo(
                "WARNING: memray module is not installed. Memray tracing will be disabled."
            )

    with profiling_context, memray_context:
        run_aea(
            ctx,
            connection_ids,
            env_file,
            is_install_deps,
            apply_environment_variables=True,
            password=password,
        )


def _calculate_connection_ids(
    ctx: Context, exclude_connections: List[PublicId]
) -> List[PublicId]:
    """Calculate resulting list of connection ids to run."""
    agent_config_manager = AgentConfigManager.load(ctx.cwd)

    exclude_connections_set = {
        connection.without_hash() for connection in exclude_connections
    }

    agent_connections = {
        connection.without_hash()
        for connection in agent_config_manager.agent_config.connections
    }
    not_existing_connections = exclude_connections_set - agent_connections

    if not_existing_connections:
        raise ValueError(
            f"Connections to exclude: {', '.join(map(str, not_existing_connections))} are not defined in agent configuration!"
        )

    connection_ids = list(agent_connections - exclude_connections_set)
    return connection_ids


@contextmanager
def _profiling_context(period: int) -> Generator:
    """Start profiling context."""
    TYPES_TO_TRACK = [
        Message,
        Dialogue,
        DialogueLabel,
        Handler,
        Model,
        Behaviour,
        Skill,
        Connection,
        Contract,
        Protocol,
    ]

    profiler = Profiling(
        period=period,
        types_to_track=TYPES_TO_TRACK,
    )
    profiler.start()
    try:
        yield None
    except Exception:  # pylint: disable=try-except-raise # pragma: nocover
        raise
    finally:
        profiler.stop()
        profiler.wait_completed(sync=True, timeout=10)
        # hack to address faulty garbage collection output being printed
        import os  # pylint: disable=import-outside-toplevel
        import sys  # pylint: disable=import-outside-toplevel

        sys.stderr = open(os.devnull, "w")


def print_table(rows: Sequence) -> None:
    """Print a formatted table."""
    head, *rows = rows
    col_lengths = list(map(len, head))

    for row in rows:
        col_lengths = list(map(max, zip(col_lengths, map(len, row))))  # type: ignore

    def _format_row(row: Tuple) -> str:
        """Format row."""
        cols = [
            s + (" " * (max(0, line - len(s)))) for line, s in zip(col_lengths, row)
        ]
        return "| " + " | ".join(cols) + " |"

    table_length = len(_format_row(head))
    separator = "=" * table_length

    click.echo(separator)
    click.echo(_format_row(head))
    click.echo(separator)
    for row in rows:
        click.echo(_format_row(row))
    click.echo(separator)
    click.echo()


def _print_instantiated_components(aea: AEA) -> None:
    """Print table of only components."""
    components: List[str] = [
        "ComponentId",
    ]

    for component_type in [
        ComponentType.PROTOCOL,
        ComponentType.CONNECTION,
        ComponentType.CONTRACT,
        ComponentType.SKILL,
    ]:
        components += [
            str(component.component_id)
            for component in aea.resources.component_registry.fetch_by_type(
                component_type
            )
        ]

    click.echo("All instantiated components")
    print_table([(c,) for c in components])


def _print_all_available_packages(ctx: Context) -> None:
    """Print hashes for all available packages"""
    ipfs_hash = IPFSHashOnly()
    rows = [("Package", "IPFSHash")]

    for package_id, package_path in list_available_packages(ctx.cwd):
        package_hash = ipfs_hash.hash_directory(str(package_path))
        rows.append((str(package_id), package_hash))

    click.echo("All available packages.")
    print_table(rows)


def _print_addresses(aea: AEA) -> None:
    """Print all the addresses used by agent."""

    addresses = [
        ("Name", "Address"),
        *((k, v) for k, v in aea.context.addresses.items()),
    ]

    click.echo("All available addresses.")
    print_table(addresses)


def run_aea(  # pylint: disable=too-many-positional-arguments
    ctx: Context,
    connection_ids: List[PublicId],
    env_file: str,
    is_install_deps: bool,
    apply_environment_variables: bool = False,
    password: Optional[str] = None,
) -> None:
    """
    Prepare and run an agent.

    :param ctx: a context object.
    :param connection_ids: list of connections public IDs.
    :param env_file: a path to env file.
    :param is_install_deps: bool flag is install dependencies.
    :param apply_environment_variables: bool flag is load environment variables.
    :param password: the password to encrypt/decrypt the private key.

    :raises ClickException: if any Exception occurs.
    """
    skip_consistency_check = ctx.config["skip_consistency_check"]
    _prepare_environment(ctx, env_file, is_install_deps)
    aea = _build_aea(
        connection_ids, skip_consistency_check, apply_environment_variables, password
    )

    click.echo(AEA_LOGO + "v" + __version__ + "\n")
    _print_all_available_packages(ctx)
    _print_instantiated_components(aea)
    _print_addresses(aea)

    click.echo(
        "Starting AEA '{}' in '{}' mode...".format(aea.name, aea.runtime.loop_mode)
    )
    try:
        aea.start()
    except KeyboardInterrupt:  # pragma: no cover
        click.echo(" AEA '{}' interrupted!".format(aea.name))  # pragma: no cover
    except Exception as e:  # pragma: no cover
        raise click.ClickException(str(e))
    finally:
        click.echo("Stopping AEA '{}' ...".format(aea.name))
        aea.stop()
        click.echo("AEA '{}' stopped.".format(aea.name))


def _prepare_environment(ctx: Context, env_file: str, is_install_deps: bool) -> None:
    """
    Prepare the AEA project environment.

    :param ctx: a context object.
    :param env_file: the path to the environment file.
    :param is_install_deps: whether to install the dependencies
    """
    load_env_file(env_file)
    if is_install_deps:
        requirements_path = REQUIREMENTS if Path(REQUIREMENTS).exists() else None
        do_install(ctx, requirement=requirements_path)


def _build_aea(
    connection_ids: Optional[List[PublicId]],
    skip_consistency_check: bool,
    apply_environment_variables: bool = False,
    password: Optional[str] = None,
) -> AEA:
    """Build the AEA."""
    try:
        builder = AEABuilder.from_aea_project(
            Path("."),
            skip_consistency_check=skip_consistency_check,
            apply_environment_variables=apply_environment_variables,
            password=password,
        )
        aea = builder.build(connection_ids=connection_ids, password=password)
        return aea
    except AEAWalletNoAddressException:
        error_msg = (
            "You haven't specified any private key for the AEA project.\n"
            "Please add one by using the commands `aea generate-key` and `aea add-key` for the ledger of your choice.\n"
        )
        raise click.ClickException(error_msg)
    except Exception as e:
        raise click.ClickException(str(e))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for subprocess task manager/tasl_test_skill."""

from aea.test_tools.test_cases import AEATestCaseEmpty


class TestTaskTestSkill(AEATestCaseEmpty):
    """Test task manager subprocess run a simple task."""

    capture_log = True

    @classmethod
    def setup_class(cls) -> None:
        """Init the test case."""
        super(TestTaskTestSkill, cls).setup_class()
        cls.add_item("skill", "fetchai/task_test_skill:0.1.0", local=True)
        cls.generate_private_key()
        cls.add_private_key()
        cls.set_config("agent.task_manager_mode", "multiprocess", "str")

    def test_task_run_in_a_subprocess(self):
        """Test task run in a subprocess."""
        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"
        assert not self.missing_from_output(
            process, ["Task set"], 10, is_terminating=False
        )
        assert not self.missing_from_output(
            process, ["result is"], 10, is_terminating=False
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Useful git tools."""

import shutil
import subprocess  # nosec


def check_working_tree_is_dirty() -> bool:
    """
    Check if the current Git working tree is dirty.

    :return: True if the working tree is not dirty, False otherwise
    """
    print("Checking whether the Git working tree is dirty...")
    result = subprocess.check_output(  # nosec
        [str(shutil.which("git")), "diff", "--stat"]
    )  # nosec
    if len(result) > 0:
        print("Git working tree is dirty:")
        print(result.decode("utf-8"))
        return False
    print("All good!")
    return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Chainlist RPC enrichment — fetch, probe, and rank public RPCs.

Optional module that downloads public RPC endpoints from chainlist.org,
validates them with ``eth_blockNumber`` probes, filters stale ones, and
returns the best candidates sorted by latency.  Used as fallback RPCs
for the RPC rotation system.
"""

# All bare except-Exception clauses in this module are intentional:
# they guard optional I/O (HTTP probes, cache reads) where any failure
# is logged and silently swallowed so the caller can degrade gracefully.
# pylint: disable=broad-exception-caught

import json
import logging
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

_logger = logging.getLogger(__name__)

# -- Constants -----------------------------------------------------------------

CHAINLIST_URL = "https://chainlist.org/rpcs.json"
CACHE_TTL = 86400  # 24 hours
MAX_CHAINLIST_CANDIDATES = 15
PROBE_TIMEOUT = 5.0  # seconds per probe
MAX_BLOCK_LAG = 10  # blocks behind median → stale
MAX_RPCS = 20  # don't enrich beyond this

_CACHE_DIR = Path(tempfile.gettempdir()) / "aea_chainlist_cache"
_CACHE_PATH = _CACHE_DIR / "rpcs.json"


# -- Helpers -------------------------------------------------------------------


def _normalize_url(url: str) -> str:
    """Lowercase + strip trailing slash for dedup comparison."""
    return url.rstrip("/").lower()


def _is_template_url(url: str) -> bool:
    """Return True if URL has template vars (requires API key)."""
    return "${" in url or "{" in url


def probe_rpc(
    url: str,
    timeout: float = PROBE_TIMEOUT,
) -> Optional[Tuple[str, float, int]]:
    """Probe *url* with ``eth_blockNumber``.

    Returns ``(url, latency_ms, block_number)`` on success, ``None`` on
    failure.

    :param url: RPC endpoint URL to probe.
    :param timeout: request timeout in seconds.
    :return: tuple of (url, latency_ms, block_number) or None on failure.
    """
    payload = json.dumps(
        {"jsonrpc": "2.0", "method": "eth_blockNumber", "params": [], "id": 1}
    ).encode()
    req = urllib.request.Request(
        url,
        data=payload,
        headers={"Content-Type": "application/json"},
    )
    try:
        start = time.monotonic()
        with urllib.request.urlopen(req, timeout=timeout) as resp:  # nosec B310
            data = json.loads(resp.read())
        latency_ms = (time.monotonic() - start) * 1000
        block_hex = data.get("result")
        if not block_hex or not isinstance(block_hex, str) or block_hex == "0x0":
            return None
        return (url, latency_ms, int(block_hex, 16))
    except Exception:  # noqa: BLE001
        return None


@dataclass
class RPCNode:
    """A single RPC entry from Chainlist."""

    url: str
    is_working: bool
    tracking: Optional[str] = None

    @property
    def is_tracking(self) -> bool:
        """True if the RPC is known to track user data."""
        return self.tracking in ("limited", "yes")


# -- Pipeline stages -----------------------------------------------------------


def _filter_candidates(
    nodes: List[RPCNode],
    existing_normalized: set,
) -> List[str]:
    """Filter to usable HTTPS candidates, dedup against existing."""
    candidates: List[str] = []
    for node in nodes:
        url = node.url
        if urllib.parse.urlparse(url).scheme != "https":
            continue
        if _is_template_url(url):
            continue
        if _normalize_url(url) in existing_normalized:
            continue
        candidates.append(url)
        if len(candidates) >= MAX_CHAINLIST_CANDIDATES:
            break
    return candidates


def _probe_candidates(candidates: List[str]) -> List[Tuple[str, float, int]]:
    """Probe candidates in parallel, return successful results."""
    results: List[Tuple[str, float, int]] = []
    n_workers = min(len(candidates), 10)
    with ThreadPoolExecutor(max_workers=n_workers) as pool:
        futures = {pool.submit(probe_rpc, url): url for url in candidates}
        for future in as_completed(futures, timeout=15):
            try:
                result = future.result()
                if result is not None:
                    results.append(result)
            except Exception:  # nosec B110  # noqa: BLE001
                pass
    return results


def _rank_and_select(
    results: List[Tuple[str, float, int]],
    candidates: List[str],
    chain_id: int,
    max_results: int,
) -> List[str]:
    """Rank by latency, filter stale RPCs (>MAX_BLOCK_LAG behind median)."""
    blocks = sorted(r[2] for r in results)
    median_block = blocks[len(blocks) // 2]

    valid = [
        (url, latency)
        for url, latency, block in results
        if median_block - block <= MAX_BLOCK_LAG
    ]
    valid.sort(key=lambda x: x[1])

    selected = [url for url, _ in valid[:max_results]]
    if selected:
        _logger.info(
            "ChainList: validated %d/%d candidates for chain %d (median block: %d)",
            len(selected),
            len(candidates),
            chain_id,
            median_block,
        )
    return selected


# -- Main class ----------------------------------------------------------------


class ChainlistRPC:
    """Fetcher and parser for Chainlist RPC data with local caching."""

    def __init__(self) -> None:
        """Initialise with empty RPC data."""
        self._data: List[Dict[str, Any]] = []

    def fetch_data(self, force_refresh: bool = False) -> None:
        """Fetch RPC data from chainlist.org (cached for 24h)."""
        if not force_refresh and _CACHE_PATH.exists():
            try:
                mtime = _CACHE_PATH.stat().st_mtime
                if time.time() - mtime < CACHE_TTL:
                    with _CACHE_PATH.open("r") as fh:
                        self._data = json.load(fh)
                    if self._data:
                        return
            except Exception:  # nosec B110  # noqa: BLE001
                pass

        try:
            req = urllib.request.Request(CHAINLIST_URL)
            with urllib.request.urlopen(req, timeout=10) as resp:  # nosec B310
                self._data = json.loads(resp.read())
            if self._data:
                _CACHE_DIR.mkdir(parents=True, exist_ok=True)
                with _CACHE_PATH.open("w") as fh:
                    json.dump(self._data, fh)
        except Exception as exc:  # noqa: BLE001
            _logger.debug("Chainlist fetch failed: %s", exc)
            if not self._data and _CACHE_PATH.exists():
                try:
                    with _CACHE_PATH.open("r") as fh:
                        self._data = json.load(fh)
                except Exception:  # nosec B110  # noqa: BLE001
                    pass
            if not self._data:
                self._data = []

    def get_rpcs(self, chain_id: int) -> List[RPCNode]:
        """Return parsed RPC nodes for *chain_id*."""
        if not self._data:
            self.fetch_data()
        for entry in self._data:
            if entry.get("chainId") == chain_id:
                return [
                    RPCNode(
                        url=rpc.get("url", ""),
                        is_working=True,
                        tracking=rpc.get("tracking"),
                    )
                    for rpc in entry.get("rpc", [])
                ]
        return []

    def get_validated_rpcs(
        self,
        chain_id: int,
        existing_rpcs: List[str],
        max_results: int = 5,
    ) -> List[str]:
        """Return Chainlist RPCs filtered, probed, and sorted by quality.

        Pipeline:
        1. Fetch HTTPS RPCs from Chainlist for *chain_id*.
        2. Filter out template URLs, duplicates, and non-HTTPS.
        3. Probe top candidates in parallel with ``eth_blockNumber``.
        4. Discard stale RPCs (block number lagging behind median).
        5. Return up to *max_results* URLs sorted by latency.

        :param chain_id: numeric EVM chain identifier.
        :param existing_rpcs: URLs already known, used for deduplication.
        :param max_results: maximum number of validated RPCs to return.
        :return: list of validated RPC URLs sorted by latency.
        """
        nodes = self.get_rpcs(chain_id)
        if not nodes:
            return []

        existing_normalized = {_normalize_url(u) for u in existing_rpcs}
        candidates = _filter_candidates(nodes, existing_normalized)
        if not candidates:
            return []

        results = _probe_candidates(candidates)
        if not results:
            return []

        return _rank_and_select(results, candidates, chain_id, max_results)


def enrich_rpc_urls(
    rpc_urls: List[str],
    chain_id: Optional[int] = None,
    max_rpcs: int = MAX_RPCS,
) -> List[str]:
    """Enrich *rpc_urls* with validated public RPCs from Chainlist.

    This is the main entry point for the RPC rotation system.
    Returns the original URLs followed by any Chainlist fallbacks.

    If *chain_id* is ``None`` or enrichment fails, returns *rpc_urls*
    unchanged.

    :param rpc_urls: existing RPC URLs to enrich.
    :param chain_id: numeric EVM chain identifier, or None to skip enrichment.
    :param max_rpcs: upper bound on total RPC URLs to return.
    :return: original URLs followed by any validated Chainlist fallbacks.
    """
    if chain_id is None or len(rpc_urls) >= max_rpcs:
        return rpc_urls

    try:
        cl = ChainlistRPC()
        extra = cl.get_validated_rpcs(
            chain_id,
            existing_rpcs=rpc_urls,
            max_results=max_rpcs - len(rpc_urls),
        )
        if extra:
            _logger.info(
                "Enriched with %d Chainlist RPCs (total: %d)",
                len(extra),
                len(rpc_urls) + len(extra),
            )
            return rpc_urls + extra
    except Exception as exc:  # pylint: disable=broad-exception-caught  # noqa: BLE001
        _logger.debug("Chainlist enrichment failed: %s", exc)

    return rpc_urls
